"""
Parsers that produces `structures.wrapper_package.WrapperPackage`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.parsers.wrapper_package_parser import WrapperPackageParser
from mypy_boto3_builder.postprocessors.aio_imports import replace_import_with_aio
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.packages.types_aioboto3_package import TypesAioBoto3Package
from mypy_boto3_builder.structures.packages.types_aiobotocore_package import TypesAioBotocorePackage
from mypy_boto3_builder.structures.packages.types_boto3_package import TypesBoto3Package
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.utils.type_checks import is_external_import


def parse_types_boto3_package(
    service_names: Iterable[ServiceName],
    package_data: BasePackageData,
    version: str,
) -> TypesBoto3Package:
    """
    Parse data for types-boto3 package.

    Arguments:
        service_names -- All available service names
        package_data -- Package data
        version -- Package version

    Returns:
        Boto3StubsPackage structure.
    """
    package = TypesBoto3Package(package_data, service_names, version)
    parser = WrapperPackageParser(package)
    package.init_functions.extend(parser.get_init_client_functions())
    package.init_functions.extend(parser.get_init_resource_functions())
    package.session_class.methods.extend(parser.get_session_client_methods())
    package.session_class.methods.extend(parser.get_session_resource_methods())
    return package


def parse_types_aiobotocore_package(
    service_names: Iterable[ServiceName],
    package_data: BasePackageData,
    version: str,
) -> TypesAioBotocorePackage:
    """
    Parse data for types-aiobotocore package.

    Arguments:
        service_names -- All available service names
        package_data -- Package data
        version -- Package version

    Returns:
        TypesAioBotocorePackage structure.
    """
    package = TypesAioBotocorePackage(package_data, service_names, version)
    parser = WrapperPackageParser(package)
    for method in parser.get_session_client_methods("create_client"):
        method.return_type = TypeSubscript(
            InternalImport("ClientCreatorContext"),
            [method.return_type],
        )
        method.type_ignore = "override"
        package.session_class.methods.append(method)

    for type_annotation in package.session_class.iterate_types():
        if is_external_import(type_annotation):
            replace_import_with_aio(type_annotation)
    return package


def parse_types_aioboto3_package(
    service_names: Iterable[ServiceName],
    package_data: BasePackageData,
    version: str,
) -> TypesAioBoto3Package:
    """
    Parse data for types-aioboto3 package.

    Arguments:
        service_names -- All available service names
        package_data -- Package data
        version -- Package version

    Returns:
        TypesAioBoto3Package structure.
    """
    package = TypesAioBoto3Package(package_data, service_names, version)
    parser = WrapperPackageParser(package)
    for method in parser.get_session_client_methods():
        method.return_type = TypeSubscript(
            ExternalImport(Import.aiobotocore + "session", "ClientCreatorContext"),
            [method.return_type],
        )
        package.session_class.methods.append(method)
    for method in parser.get_session_resource_methods():
        method.return_type = TypeSubscript(
            InternalImport("ResourceCreatorContext"),
            [method.return_type],
        )
        package.session_class.methods.append(method)

    for type_annotation in package.session_class.iterate_types():
        if is_external_import(type_annotation):
            replace_import_with_aio(type_annotation)
    return package



"""
Base stubs/docs generator.

Copyright 2024 Vlad Emelianov
"""

import shutil
import tempfile
from abc import ABC, abstractmethod
from collections.abc import Sequence
from pathlib import Path
from typing import ClassVar

from mypy_boto3_builder.cli_parser import CLINamespace
from mypy_boto3_builder.enums.product import Product
from mypy_boto3_builder.enums.product_type import ProductType
from mypy_boto3_builder.exceptions import AlreadyPublishedError
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.parsers.service_package_parser import ServicePackageParser
from mypy_boto3_builder.postprocessors.base import BasePostprocessor
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.package_extra import PackageExtra
from mypy_boto3_builder.structures.packages.service_package import ServicePackage
from mypy_boto3_builder.utils.github import download_and_extract
from mypy_boto3_builder.utils.package_builder import PackageBuilder
from mypy_boto3_builder.utils.pypi_manager import PyPIManager
from mypy_boto3_builder.writers.package_writer import PackageWriter


class BaseGenerator(ABC):
    """
    Base stubs/docs generator.

    Arguments:
        service_names -- Selected service names
        main_service_names -- Service names included in main
        config -- CLI configuration
        version -- Package build version
        cleanup -- Whether to cleanup generated files
    """

    service_package_data: ClassVar[BasePackageData]
    service_template_path: ClassVar[Path]

    def __init__(
        self,
        product: Product,
        service_names: Sequence[ServiceName],
        main_service_names: Sequence[ServiceName],
        config: CLINamespace,
        version: str,
        *,
        cleanup: bool,
    ) -> None:
        self.__temp_path: Path | None = None
        self._cleanup_dirs: list[Path] = []
        self._downloaded_static_files_path: Path | None = None

        self.product = product
        self.service_names = service_names
        self.main_service_names = main_service_names
        self.config = config
        self.logger = get_logger()
        self.version = version or self._get_library_version()
        self.cleanup = cleanup
        self.package_writer = PackageWriter(
            output_path=self.output_path,
            generate_package=self.is_package(),
            cleanup=cleanup,
        )
        self.setup_package_writer = PackageWriter(
            output_path=self.output_path,
            generate_package=self.is_package(),
            cleanup=False,
        )

    def is_package(self) -> bool:
        """
        Whether to generate ready-to-install package or installed package.
        """
        if self.product.get_type() == ProductType.docs:
            return False
        return not any(output_type.is_installed() for output_type in self.config.output_types)

    def is_packaged(self) -> bool:
        """
        Whether to build wheel or sdist.
        """
        return any(output_type.is_packaged() for output_type in self.config.output_types)

    def is_package_temporary(self) -> bool:
        """
        Whether to preserve generated package.
        """
        return not any(output_type.is_preserved() for output_type in self.config.output_types)

    @property
    def _temp_path(self) -> Path:
        if not self.__temp_path:
            self.__temp_path = Path(tempfile.TemporaryDirectory(delete=False).name)
            self._cleanup_dirs.append(self.__temp_path)
        return self.__temp_path

    @property
    def output_path(self) -> Path:
        """
        Output path.
        """
        if self.is_package_temporary():
            return self._temp_path
        return self.config.output_path

    def _get_or_download_static_files_path(
        self,
        static_path: Path,
        download_url: str | None = None,
    ) -> Path:
        """
        Get path to static files. Download if needed.
        """
        if not self.config.download_static_stubs or not download_url:
            return static_path

        if self._downloaded_static_files_path:
            return self._downloaded_static_files_path

        self.logger.debug(f"Downloading static files from {download_url}")
        temp_dir_path = Path(tempfile.TemporaryDirectory(delete=False).name)
        self._cleanup_dirs.append(temp_dir_path)
        self._downloaded_static_files_path = download_and_extract(download_url, temp_dir_path)

        self.logger.debug(f"Downloaded static files to {self._downloaded_static_files_path}")

        return self._downloaded_static_files_path

    @abstractmethod
    def _get_postprocessor(self, service_package: ServicePackage) -> BasePostprocessor:
        """
        Get postprocessor for service package.
        """

    def _get_library_version(self) -> str:
        """
        Get underlying library version.
        """
        return self.service_package_data.get_library_version()

    def _get_package_build_version(self, pypi_name: str) -> str:
        if self.config.disable_smart_version:
            return self.version
        pypi_manager = PyPIManager(pypi_name)
        if not pypi_manager.has_version(self.version):
            return self.version

        if self.config.skip_published:
            self.logger.info(
                f"Skipping {pypi_name}: {self.version} is already on PyPI",
                tags=pypi_name,
            )
            raise AlreadyPublishedError(f"{pypi_name} {self.version} is already on PyPI")

        return pypi_manager.get_next_version(self.version)

    @abstractmethod
    def generate_stubs(self) -> Package | None:
        """
        Generate main stubs.
        """

    @abstractmethod
    def generate_stubs_lite(self) -> Package | None:
        """
        Generate main stubs.
        """

    @abstractmethod
    def generate_full_stubs(self) -> Package | None:
        """
        Generate full stubs.
        """

    @abstractmethod
    def generate_custom_stubs(self) -> Package | None:
        """
        Generate custom stubs.
        """

    def _generate_full_stubs_services(self, package: Package) -> None:
        service_package_writer = PackageWriter(
            output_path=self.output_path / package.directory_name,
            generate_package=False,
            cleanup=False,
        )
        total_str = f"{len(self.service_names)}"
        for index, service_name in enumerate(self.service_names):
            current_str = f"{{:0{len(total_str)}}}".format(index + 1)
            progress_str = f"[{current_str}/{total_str}]"

            self.logger.info(
                f"{progress_str} Generating {service_name.boto3_name} service directory",
                tags=service_name.boto3_name,
            )
            service_package = self._parse_service_package(
                service_name=service_name,
                version=package.version,
                package_data=package.data,
            )
            service_package.mark_safe_typed_dicts()

            service_package.pypi_name = package.pypi_name
            service_package_writer.write_service_package(
                package=service_package,
                templates_path=self.service_template_path,
            )
            package.install_requires.update(service_package.install_requires)

    @abstractmethod
    def generate_docs(self) -> None:
        """
        Generate service and main docs.
        """

    def generate_product(self, product_type: ProductType) -> None:
        """
        Run generator for a product type.
        """
        packages: list[Package | None] = []
        match product_type:
            case ProductType.stubs:
                packages.append(self.generate_stubs())
            case ProductType.stubs_lite:
                packages.append(self.generate_stubs_lite())
            case ProductType.service_stubs:
                packages.extend(self.generate_service_stubs())
            case ProductType.docs:
                self.generate_docs()
            case ProductType.full:
                packages.append(self.generate_full_stubs())
            case ProductType.custom:
                packages.append(self.generate_custom_stubs())

        generated_packages = list(filter(None, packages))
        if self.is_packaged() and generated_packages:
            package_builder = PackageBuilder(
                build_path=self.output_path,
                output_path=self.config.output_path,
            )
            for package in generated_packages:
                package_builder.build(package, self.config.output_types)

    def _parse_service_package(
        self,
        service_name: ServiceName,
        version: str,
        package_data: BasePackageData,
    ) -> ServicePackage:
        self.logger.debug(
            f"Parsing {service_name.boto3_name} botocore service",
            tags=service_name.boto3_name,
        )
        parser = ServicePackageParser(service_name, package_data, version)
        service_package = parser.parse()

        postprocessor = self._get_postprocessor(service_package)
        postprocessor.generate_docstrings()
        postprocessor.process_package()

        postprocessor.extend_literals()
        postprocessor.replace_self_ref_typed_dicts()
        return service_package

    def _process_service(
        self,
        service_name: ServiceName,
        version: str,
        package_data: BasePackageData,
        templates_path: Path,
    ) -> ServicePackage:
        service_package = self._parse_service_package(
            service_name=service_name,
            version=version,
            package_data=package_data,
        )
        service_package.mark_safe_typed_dicts()

        self.logger.debug(
            f"Writing {service_name.boto3_name} service package",
            tags=service_name.boto3_name,
        )
        self.package_writer.write_service_package(
            service_package,
            templates_path=templates_path,
        )
        return service_package

    def _process_service_docs(
        self,
        service_name: ServiceName,
        version: str,
        package_data: BasePackageData,
        templates_path: Path,
    ) -> ServicePackage:
        service_package = self._parse_service_package(
            service_name=service_name,
            version=version,
            package_data=package_data,
        )

        self.logger.debug(
            f"Writing {service_name.boto3_name} service docs",
            tags=service_name.boto3_name,
        )
        self.package_writer.write_service_docs(
            service_package,
            templates_path=templates_path,
        )
        return service_package

    def generate_service_stubs(self) -> list[ServicePackage]:
        """
        Generate service stubs.
        """
        total_str = f"{len(self.service_names)}"
        packages: list[ServicePackage] = []
        for index, service_name in enumerate(self.service_names):
            current_str = f"{{:0{len(total_str)}}}".format(index + 1)
            progress_str = f"[{current_str}/{total_str}]"

            pypi_name = self.service_package_data.get_service_pypi_name(service_name)
            try:
                version = self._get_package_build_version(pypi_name)
            except AlreadyPublishedError:
                continue

            self.logger.info(f"{progress_str} Generating {pypi_name} {version}", tags=pypi_name)
            package = self._process_service(
                service_name=service_name,
                version=version,
                package_data=self.service_package_data,
                templates_path=self.service_template_path,
            )
            packages.append(package)

        return packages

    def cleanup_temporary_files(self) -> None:
        """
        Cleanup temporary files.
        """
        for dir_path in self._cleanup_dirs:
            if not dir_path.exists():
                continue

            self.logger.debug(f"Removing {dir_path}")
            shutil.rmtree(dir_path)

    def _get_wrapper_package_extras(self, package: Package) -> tuple[PackageExtra, ...]:
        result: list[PackageExtra] = []
        if package.data.pypi_full_name:
            result.append(
                PackageExtra(
                    ServiceName.FULL,
                    (
                        f"{package.data.pypi_full_name}"
                        f">={package.min_library_version}, <{package.max_library_version}",
                    ),
                ),
            )
        result.append(
            PackageExtra(
                package.library_name,
                (f"{package.library_name}=={package.library_version}",),
            )
        )
        if package.service_names:
            result.append(
                PackageExtra(
                    ServiceName.ALL,
                    tuple(
                        (
                            f"{package.data.get_service_pypi_name(service_name)}"
                            f">={package.min_library_version}, <{package.max_library_version}"
                        )
                        for service_name in package.service_names
                    ),
                ),
            )
        if package.essential_service_names:
            result.append(
                PackageExtra(
                    ServiceName.ESSENTIAL,
                    tuple(
                        (
                            f"{package.data.get_service_pypi_name(service_name)}"
                            f">={package.min_library_version}, <{package.max_library_version}"
                        )
                        for service_name in package.essential_service_names
                    ),
                ),
            )
        result.extend(
            PackageExtra(
                service_name.extras_name,
                (
                    f"{package.data.get_service_pypi_name(service_name)}"
                    f">={package.min_library_version}, <{package.max_library_version}",
                ),
            )
            for service_name in package.service_names
        )
        return tuple(result)

    def _log_generate(self, pypi_name: str, version: str | None = None) -> None:
        version_str = version or self.version
        self.logger.info(
            f"Generating {pypi_name} {version_str}",
            tags=pypi_name,
        )



"""
Constants and paths.

Copyright 2024 Vlad Emelianov
"""

from pathlib import Path
from typing import Final

# Root path for the project
ROOT_PATH: Final = Path(__file__).parent

# Jinja2 templates
TEMPLATES_PATH: Final = ROOT_PATH / "templates"

# Max line length for formatting
LINE_LENGTH: Final = 100

# Max line length for formatting docstrings
DOCSTRING_LINE_LENGTH: Final = 80

# Max docstring length extracted from method
DOCSTRING_MAX_LENGTH: Final = 300

# mypy-boto3-builder GitHub link
BUILDER_REPO_URL: Final = "https://github.com/youtype/mypy_boto3_builder"

# Main logger name
LOGGER_NAME: Final = "mypy_boto3_builder"

# builder CLI entrypoint name
PROG_NAME: Final = "mypy_boto3_builder"

# builder package name
PACKAGE_NAME: Final = "mypy-boto3-builder"

# universal mask for all resources
ALL: Final = "*"
ATTRIBUTES: Final = "_attributes"
SERVICE_RESOURCE: Final = "ServiceResource"
CLIENT: Final = "Client"

TYPING_EXTENSIONS_PYPI_NAME: Final = "typing-extensions"

# python versions supported by output stubs
SUPPORTED_PY_VERSIONS: Final = {
    (3, 8),
    (3, 9),
    (3, 10),
    (3, 11),
    (3, 12),
    (3, 13),
    (3, 14),
}

# default timeout for HTTP requests
REQUEST_TIMEOUT: Final = 120

# Sentinel value for output path for CLI
OUTPUT_PATH_SENTINEL = Path("/tmp/output_path_sentinel")  # noqa: S108


class TemplatePath:
    """
    Template paths.
    """

    mypy_boto3: Final = TEMPLATES_PATH / "mypy-boto3"

    types_boto3: Final = TEMPLATES_PATH / "types-boto3"
    types_boto3_docs: Final = TEMPLATES_PATH / "types-boto3-docs"
    types_boto3_service: Final = TEMPLATES_PATH / "types-boto3-service"
    types_boto3_service_docs: Final = TEMPLATES_PATH / "types-boto3-service-docs"
    types_boto3_full: Final = TEMPLATES_PATH / "types-boto3-full"
    types_boto3_custom: Final = TEMPLATES_PATH / "types-boto3-custom"

    types_aioboto3: Final = TEMPLATES_PATH / "types-aioboto3"
    types_aioboto3_docs: Final = TEMPLATES_PATH / "types-aioboto3-docs"
    types_aioboto3_service_docs: Final = TEMPLATES_PATH / "types-aioboto3-service-docs"
    types_aioboto3_custom: Final = TEMPLATES_PATH / "types-aioboto3-custom"

    types_aiobotocore: Final = TEMPLATES_PATH / "types-aiobotocore"
    types_aiobotocore_docs: Final = TEMPLATES_PATH / "types-aiobotocore-docs"
    types_aiobotocore_service: Final = TEMPLATES_PATH / "types-aiobotocore-service"
    types_aiobotocore_service_docs: Final = TEMPLATES_PATH / "types-aiobotocore-service-docs"
    types_aiobotocore_full: Final = TEMPLATES_PATH / "types-aiobotocore-full"
    types_aiobotocore_custom: Final = TEMPLATES_PATH / "types-aiobotocore-custom"


class StaticStubsPath:
    """
    Static *.pyi files path.
    """

    types_boto3: Final = ROOT_PATH / "stubs_static" / "types-boto3"
    types_aiobotocore: Final = ROOT_PATH / "stubs_static" / "types-aiobotocore"
    types_aioboto3: Final = ROOT_PATH / "stubs_static" / "types-aioboto3"


class StaticStubsPullURL:
    """
    Static *.pyi files zip download URL.
    """

    types_boto3: Final = "https://api.github.com/repos/youtype/types-boto3/zipball/main"
    types_aiobotocore: Final = "https://api.github.com/repos/youtype/types-aiobotocore/zipball/main"
    types_aioboto3: Final = "https://api.github.com/repos/youtype/types-aioboto3/zipball/main"



"""
Type annotations for aioboto3.resources.base module.

Copyright 2025 Vlad Emelianov
"""

import logging
from types import TracebackType
from typing import TypeVar

from boto3.resources.base import ServiceResource

_R = TypeVar("_R", bound=AIOBoto3ServiceResource)

logger: logging.Logger

class AIOBoto3ServiceResource(ServiceResource):
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...
    def close(self) -> None: ...



"""
Parser for botocore Client, produces `structures.Client`.

Copyright 2025 Vlad Emelianov
"""

from botocore.client import ClientMeta
from botocore.errorfactory import ClientExceptionsFactory
from botocore.exceptions import ClientError

from mypy_boto3_builder.constants import CLIENT
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.client import Client
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.service_stub_map import get_stub_method_map
from mypy_boto3_builder.utils.strings import get_class_prefix


class ClientParser:
    """
    Parser for botocore Client, produces `structures.Client`.

    Arguments:
        service_name -- Target service name.
        shape_parser -- ShapeParser instance.
    """

    def __init__(
        self,
        service_name: ServiceName,
        shape_parser: ShapeParser,
    ) -> None:
        self.name = CLIENT
        self.alias = service_name.get_client_name()
        self.service_name = service_name
        self.shape_parser = shape_parser
        self._logger = get_logger()

    def _parse_method_map(self) -> dict[str, Method]:
        shape_method_map = self.shape_parser.get_client_method_map()
        stub_method_map = get_stub_method_map(self.service_name, self.name)
        for method in stub_method_map.values():
            self.shape_parser.create_request_type_annotation(
                method=method,
                name=f"{self.name}{get_class_prefix(method.name)}Request",
            )
        return {**shape_method_map, **stub_method_map}

    def _parse_exception_class_attributes(self) -> list[Attribute]:
        result: list[Attribute] = []
        client_exceptions = ClientExceptionsFactory().create_client_exceptions(
            self.shape_parser.service_model
        )
        for exception_class_name in dir(client_exceptions):
            if exception_class_name.startswith("_"):
                continue
            if not exception_class_name[0].isupper():
                continue
            result.append(
                Attribute(
                    exception_class_name,
                    TypeSubscript(
                        Type.Type,
                        [ExternalImport.from_class(ClientError, alias="BotocoreClientError")],
                    ),
                ),
            )
        return result

    def parse(self) -> Client:
        """
        Parse botocore Client.

        Returns:
            Client structure.
        """
        self._logger.debug(f"Parsing {self.alias}", tags=self.alias)

        result = Client(
            name=self.alias,
            service_name=self.service_name,
        )

        method_map = self._parse_method_map()
        result.methods.append(result.get_exceptions_property())
        result.methods.extend(method_map.values())

        result.attributes.append(Attribute("meta", ExternalImport.from_class(ClientMeta)))

        result.exceptions_class.attributes.extend(self._parse_exception_class_attributes())

        return result



"""
PyPI package data constants.

Copyright 2024 Vlad Emelianov
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass

from mypy_boto3_builder.enums.product_library import ProductLibrary
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.utils.version_getters import (
    get_aioboto3_version,
    get_aiobotocore_version,
    get_boto3_version,
)


@dataclass(kw_only=True)
class BasePackageData(ABC):
    """
    Generic package data.
    """

    # package name
    name: str

    # package name on PyPI
    pypi_name: str

    # main stubs package name on PyPI
    pypi_stubs_name: str

    # lite stubs package name on PyPI
    pypi_lite_name: str

    # full stubs package name on PyPI
    pypi_full_name: str

    # custom stubs package name on PyPI
    pypi_custom_name: str

    # underlying library: boto3, aiobotocore, aioboto3
    library: ProductLibrary

    # service package name prefix
    service_prefix: str

    # service package name prefix on PyPI
    service_pypi_prefix: str

    # link to local documentation
    local_doc_link: str

    # install_requires for setup.py
    install_requires: tuple[str, ...]

    # is package installable with VSCode extensions
    is_vscode_supported: bool = False

    # is the package released to conda-forge
    is_conda_forge_supported: bool = False

    def get_service_package_name(self, service_name: ServiceName) -> str:
        """
        Get service package name.
        """
        return f"{self.service_prefix}_{service_name.underscore_name}"

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return f"{self.service_pypi_prefix}-{service_name.name}"

    def get_library_version(self) -> str:
        """
        Get underlying library version.
        """
        return self._get_library_version()

    @abstractmethod
    def _get_library_version(self) -> str:
        raise NotImplementedError("Method not implemented")

    def has_pypi_lite_package(self) -> bool:
        """
        Check if package has lite version.
        """
        return bool(self.pypi_lite_name and self.pypi_lite_name != self.pypi_name)

    def get_library_name(self) -> str:
        """
        Get PyPI library name.
        """
        return self.library.get_library_name()

    def get_library_chat_choice(self) -> str:
        """
        Get package library chat choice.
        """
        return self.library.get_chat_choice()


@dataclass(kw_only=True)
class TypesBoto3PackageData(BasePackageData):
    """
    types-boto3 package data.
    """

    name: str = "boto3-stubs"
    pypi_name: str = "types-boto3"
    pypi_stubs_name: str = "types-boto3"
    pypi_lite_name: str = "types-boto3-lite"
    pypi_full_name: str = "types-boto3-full"
    pypi_custom_name: str = "types-boto3-custom"
    library: ProductLibrary = ProductLibrary.boto3
    service_prefix: str = "types_boto3"
    service_pypi_prefix: str = "types-boto3"
    local_doc_link: str = "https://youtype.github.io/types_boto3_docs/"
    install_requires: tuple[str, ...] = (
        "botocore-stubs",
        "types-s3transfer",
        'typing-extensions>=4.1.0; python_version<"3.12"',
    )
    is_vscode_supported: bool = True
    is_conda_forge_supported: bool = False

    def _get_library_version(self) -> str:
        return get_boto3_version()


@dataclass(kw_only=True)
class TypesBoto3LitePackageData(TypesBoto3PackageData):
    """
    types-boto3-lite package data.
    """

    pypi_name: str = TypesBoto3PackageData.pypi_lite_name
    pypi_lite_name: str = ""


@dataclass(kw_only=True)
class TypesBoto3FullPackageData(TypesBoto3PackageData):
    """
    boto3-stubs-full package data.
    """

    name: str = ""
    pypi_name: str = TypesBoto3PackageData.pypi_full_name
    is_conda_forge_supported: bool = False
    install_requires: tuple[str, ...] = ()

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name


@dataclass(kw_only=True)
class TypesBoto3CustomPackageData(TypesBoto3PackageData):
    """
    boto3-stubs-custom package data.
    """

    pypi_name: str = "types-boto3-custom"
    is_conda_forge_supported: bool = False

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name


@dataclass(kw_only=True)
class TypesAioBotocorePackageData(BasePackageData):
    """
    types-aiobotocore package data.
    """

    name: str = "aiobotocore-stubs"
    pypi_name: str = "types-aiobotocore"
    pypi_stubs_name: str = "types-aiobotocore"
    pypi_lite_name: str = "types-aiobotocore-lite"
    pypi_full_name: str = "types-aiobotocore-full"
    pypi_custom_name: str = "types-aiobotocore-custom"
    library: ProductLibrary = ProductLibrary.aiobotocore
    service_prefix: str = "types_aiobotocore"
    service_pypi_prefix: str = "types-aiobotocore"
    local_doc_link: str = "https://youtype.github.io/types_aiobotocore_docs/"
    install_requires: tuple[str, ...] = (
        "botocore-stubs",
        'typing-extensions>=4.1.0; python_version<"3.12"',
    )

    def _get_library_version(self) -> str:
        return get_aiobotocore_version()


@dataclass(kw_only=True)
class TypesAioBotocoreLitePackageData(TypesAioBotocorePackageData):
    """
    types-aiobotocore-lite package data.
    """

    pypi_name: str = TypesAioBotocorePackageData.pypi_lite_name
    pypi_lite_name: str = ""


@dataclass(kw_only=True)
class TypesAioBotocoreFullPackageData(TypesAioBotocorePackageData):
    """
    types-aiobotocore-full package data.
    """

    name: str = ""
    pypi_name: str = TypesAioBotocorePackageData.pypi_full_name
    is_conda_forge_supported: bool = False
    install_requires: tuple[str, ...] = ()

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name


@dataclass(kw_only=True)
class TypesAioBotocoreCustomPackageData(TypesAioBotocorePackageData):
    """
    types-aiobotocore-custom package data.
    """

    pypi_name: str = TypesAioBotocorePackageData.pypi_custom_name
    is_conda_forge_supported: bool = False

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name


@dataclass(kw_only=True)
class Boto3StubsPackageData(BasePackageData):
    """
    boto3-stubs package data.
    """

    name: str = "boto3-stubs"
    pypi_name: str = "boto3-stubs"
    pypi_stubs_name: str = "boto3-stubs"
    pypi_lite_name: str = "boto3-stubs-lite"
    pypi_full_name: str = "boto3-stubs-full"
    pypi_custom_name: str = "boto3-stubs-custom"
    service_prefix: str = "mypy_boto3"
    service_pypi_prefix: str = "mypy-boto3"
    library: ProductLibrary = ProductLibrary.boto3_legacy
    local_doc_link: str = "https://youtype.github.io/boto3_stubs_docs/"
    install_requires: tuple[str, ...] = (
        "botocore-stubs",
        "types-s3transfer",
        'typing-extensions>=4.1.0; python_version<"3.12"',
    )
    is_vscode_supported: bool = True
    is_conda_forge_supported: bool = True

    def _get_library_version(self) -> str:
        return get_boto3_version()


@dataclass(kw_only=True)
class Boto3StubsLitePackageData(Boto3StubsPackageData):
    """
    boto3-stubs-lite package data.
    """

    pypi_name: str = Boto3StubsPackageData.pypi_lite_name
    pypi_lite_name: str = ""


@dataclass(kw_only=True)
class Boto3StubsFullPackageData(Boto3StubsPackageData):
    """
    boto3-stubs-full package data.
    """

    name: str = ""
    pypi_name: str = Boto3StubsPackageData.pypi_full_name
    is_conda_forge_supported: bool = False
    install_requires: tuple[str, ...] = ()

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name


@dataclass(kw_only=True)
class Boto3StubsCustomPackageData(Boto3StubsPackageData):
    """
    boto3-stubs-custom package data.
    """

    pypi_name: str = Boto3StubsPackageData.pypi_custom_name
    is_conda_forge_supported: bool = False

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name


@dataclass(kw_only=True)
class MypyBoto3PackageData(Boto3StubsPackageData):
    """
    mypy-boto3 package data.
    """

    name: str = "mypy_boto3"
    pypi_name: str = "mypy-boto3"
    library: ProductLibrary = ProductLibrary.boto3_legacy
    is_vscode_supported: bool = False
    is_conda_forge_supported: bool = False


@dataclass(kw_only=True)
class TypesAioBoto3PackageData(BasePackageData):
    """
    types-aioboto3 package data.
    """

    name: str = "aioboto3-stubs"
    pypi_name: str = "types-aioboto3"
    pypi_stubs_name: str = "types-aioboto3"
    pypi_lite_name: str = "types-aioboto3-lite"
    pypi_full_name: str = TypesAioBotocorePackageData.pypi_full_name
    pypi_custom_name: str = "types-aioboto3-custom"
    library: ProductLibrary = ProductLibrary.aioboto3
    service_prefix: str = "types_aiobotocore"
    service_pypi_prefix: str = "types-aiobotocore"
    local_doc_link: str = "https://youtype.github.io/types_aioboto3_docs/"
    install_requires: tuple[str, ...] = (
        "botocore-stubs",
        TypesAioBotocorePackageData.pypi_stubs_name,
        "types-s3transfer",
        'typing-extensions>=4.1.0; python_version<"3.12"',
    )

    def _get_library_version(self) -> str:
        return get_aioboto3_version()


@dataclass(kw_only=True)
class TypesAioBoto3LitePackageData(TypesAioBoto3PackageData):
    """
    types-aioboto3-lite package data.
    """

    pypi_name: str = TypesAioBoto3PackageData.pypi_lite_name
    pypi_lite_name: str = ""
    install_requires: tuple[str, ...] = (
        "botocore-stubs",
        TypesAioBotocorePackageData.pypi_lite_name,
        "types-s3transfer",
        'typing-extensions>=4.1.0; python_version<"3.12"',
    )


@dataclass(kw_only=True)
class TypesAioBoto3CustomPackageData(TypesAioBoto3PackageData):
    """
    types-aioboto3-custom package data.
    """

    pypi_name: str = TypesAioBoto3PackageData.pypi_custom_name
    is_conda_forge_supported: bool = False
    install_requires: tuple[str, ...] = (
        "botocore-stubs",
        "types-s3transfer",
        'typing-extensions>=4.1.0; python_version<"3.12"',
    )

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get service package PyPI name.
        """
        return self.pypi_name



"""
Interactive chat buddy to help user to select services and generate type annotations.

Copyright 2024 Vlad Emelianov
"""

import logging
import sys
import time
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Final

from mypy_boto3_builder.chat.chat import Chat, Choice
from mypy_boto3_builder.chat.enums import Library, PackageManager, ServiceActions
from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message, MessageToken
from mypy_boto3_builder.cli_parser import CLINamespace
from mypy_boto3_builder.constants import PROG_NAME
from mypy_boto3_builder.enums.output_type import OutputType
from mypy_boto3_builder.logger import enable_logger
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.utils.boto3_utils import get_available_service_names
from mypy_boto3_builder.utils.path import print_path
from mypy_boto3_builder.utils.version_getters import get_botocore_version

MAX_SERVICE_PRINTED = 20
MAX_SERVICE_PYCHARM = 20
DEFAULT_OUTPUT_PATH = Path("./vendored")
REPORT_URL = "https://github.com/youtype/mypy_boto3_builder/issues"
PAIR = 2
PROJECT_PATH = Path.cwd()


def _linebreak() -> None:
    sys.stdout.write("\n")


def _join_and(items: Sequence[MessageToken], name: str = "") -> Message:
    name_suffix = f" {name}{'' if len(items) == 1 else 's'}" if name else ""
    if not items:
        return ("no ", name_suffix)
    if len(items) == 1:
        return ("only ", items[0], name_suffix)
    if len(items) == PAIR:
        return (items[0], " and ", items[1], name_suffix)

    result: list[MessageToken] = []
    for item in items[:-1]:
        result.extend((item, ", "))

    result.extend(("and ", items[-1], name_suffix))
    return result


class ChatBuddy:
    """
    Interactive chat buddy to help user to select services and generate type annotations.
    """

    START_SHORTCUT_KEY = "0"
    SERVICE_SELECT_HELP: Final[Message] = (
        "Type ",
        TextStyle.tag.wrap("first letter"),
        " to search, type ",
        TextStyle.tag.wrap(START_SHORTCUT_KEY),
        " to return to the top.",
    )

    def __init__(
        self,
        run_builder: Callable[[CLINamespace], None],
        external_args: CLINamespace,
    ) -> None:
        self.available_service_names: tuple[ServiceName, ...] = ()
        self.selected_service_names: list[ServiceName] = []
        self.recommended_service_names: list[ServiceName] = []
        self._library: Library | None = Library.from_products(external_args.products)
        self.output_path: Path = DEFAULT_OUTPUT_PATH
        self.project_path = PROJECT_PATH
        self.args = CLINamespace(
            log_level=logging.ERROR,
            output_path=self.output_path,
            service_names=[],
            build_version="",
            output_types=[OutputType.wheel],
            products=[],
            disable_smart_version=False,
            download_static_stubs=True,
        )
        self.package_manager: PackageManager = PackageManager.pip
        self.run_builder = run_builder
        self.chat = Chat()

    @property
    def library(self) -> Library:
        """
        Get selected library.
        """
        if self._library is None:
            raise ValueError("Library is not selected")

        return self._library

    @property
    def library_name(self) -> str:
        """
        Get library name string.
        """
        return self.library.product_library.get_library_name()

    def _get_selected_service_names(self) -> list[ServiceName]:
        return [i for i in self.available_service_names if i.is_essential()]

    def _say(self, *messages: Message | str) -> None:
        for message in messages[:-1]:
            self.chat.say(message)
            time.sleep(0.3)

        self.chat.say(messages[-1])

    def _respond(self, *messages: Message | str) -> None:
        for message in messages:
            self.chat.respond(message)

    def _get_service_choices(
        self,
        service_names: Sequence[ServiceName],
        selected_suffix: str,
        selected: Iterable[ServiceName] = (),
    ) -> list[Choice]:
        sorted_service_names = sorted(service_names, key=lambda x: x.class_name)
        last_shortcut = None
        result: list[Choice] = []
        for service_name in sorted_service_names:
            current_shortcut = service_name.class_name[0].lower()
            shortcut_key: str | None = None
            if current_shortcut != last_shortcut:
                last_shortcut = current_shortcut
                shortcut_key = current_shortcut

            is_selected = service_name in selected
            choice_selected_suffix = (
                TextStyle.tag.apply(selected_suffix)
                if is_selected
                else TextStyle.hilight.apply(selected_suffix)
            )
            choice = Choice(
                title=(
                    service_name.class_name,
                    TextStyle.dim.wrap(f" ({service_name.boto3_name})"),
                ),
                key=service_name.class_name,
                shortcut_key=shortcut_key,
                selected_suffix=choice_selected_suffix,
                disabled=is_selected,
            )
            if is_selected:
                choice.select()
            result.append(choice)
        return result

    def _add_service_names(self) -> None:
        choice_service_names = [
            i for i in self.available_service_names if i not in self.selected_service_names
        ]
        service_choices = self._get_service_choices(
            service_names=self.available_service_names,
            selected=set(self.selected_service_names),
            selected_suffix=" ✓",
        )
        service_choice_map = {i.class_name: i for i in choice_service_names}
        selected = self._select_multiple(
            message="I use",
            choices=service_choices,
            finish_choice=Choice(
                title=["Nothing else", TextStyle.dim.wrap(" (continue)")],
                text="nothing else",
                shortcut_key=self.START_SHORTCUT_KEY,
            ),
            finish_selected_text="",
            instruction=self.SERVICE_SELECT_HELP,
        )
        selected_services = [service_choice_map[i] for i in selected]
        self.selected_service_names.extend(selected_services)

    def _remove_service_names(self) -> None:
        service_choices = self._get_service_choices(
            service_names=self.selected_service_names,
            selected_suffix=" ✗",
        )
        service_choice_map = {i.class_name: i for i in self.selected_service_names}
        selected = self._select_multiple(
            message="I do not use",
            choices=service_choices,
            finish_choice=Choice(
                title=["Anything else", TextStyle.dim.wrap(" (continue)")],
                text="anything else",
                shortcut_key=self.START_SHORTCUT_KEY,
            ),
            finish_selected_text="",
            instruction=self.SERVICE_SELECT_HELP,
        )
        selected_services = [service_choice_map[i] for i in selected]
        for selected_service_name in selected_services:
            self.selected_service_names.remove(selected_service_name)

    def _is_all_selected(self) -> bool:
        return len(self.selected_service_names) == len(self.available_service_names)

    def _get_selected_services_message(self) -> Message:
        selected = self.selected_service_names
        if not selected:
            return (TextStyle.tag.wrap("no"), " services")
        if self._is_all_selected():
            return (
                TextStyle.tag.wrap("all"),
                " ",
                TextStyle.tag.wrap(len(selected)),
                " available services",
            )
        if len(selected) > MAX_SERVICE_PRINTED:
            selected_strs = [
                *(TextStyle.tag.wrap(i.class_name) for i in selected[: MAX_SERVICE_PRINTED - 1]),
                TextStyle.tag.wrap(f"{len(selected) - MAX_SERVICE_PRINTED + 1} other"),
            ]
        else:
            selected_strs = [TextStyle.tag.wrap(i.class_name) for i in selected]
        return _join_and(selected_strs, "service")

    def _get_services_messages(self) -> list[Message]:
        selected = self.selected_service_names
        result: list[Message] = []
        if len(self.selected_service_names) > MAX_SERVICE_PYCHARM:
            result.append(
                (
                    "If you use ",
                    TextStyle.tag.wrap("PyCharm"),
                    " select less than ",
                    TextStyle.tag.wrap(MAX_SERVICE_PYCHARM),
                    " services. Otherwise, you may have peroformance issues.",
                )
            )
        if self._is_all_selected():
            result.extend(
                (
                    (
                        "Do you really need type checking for ",
                        TextStyle.tag.wrap("ALL"),
                        " available services?",
                    ),
                    (
                        "Building all takes ",
                        TextStyle.tag.wrap("5-10 minutes"),
                        ", and package size is around ",
                        TextStyle.tag.wrap("12 megabytes"),
                        "!",
                    ),
                )
            )
            return result
        if not selected:
            result.append(("Okay, what service do you want to ", TextStyle.tag.wrap("add"), "?"))
            return result

        if len(selected) == 1:
            result.append(
                (
                    "Should I add type checking only for ",
                    TextStyle.tag.wrap(selected[0].class_name),
                    " service?",
                )
            )
            return result

        result.append(("Do you use ", *self._get_selected_services_message(), "?"))
        return result

    def _select(
        self,
        message: str,
        choices: Sequence[Choice | str],
        *,
        default: str | None = None,
        message_end: str = ".",
        instruction: Message | str = "",
    ) -> str:
        result = self.chat.select(
            message=message,
            choices=choices,
            default=default,
            message_end=message_end,
            instruction=instruction,
        )
        _linebreak()
        return result

    def _select_multiple(
        self,
        message: str,
        choices: Sequence[Choice | str],
        *,
        default: str | None = None,
        message_end: str = ".",
        instruction: Message | str = "",
        finish_choice: Choice | str = "Go back",
        finish_selected_text: Message | str | None = None,
    ) -> list[str]:
        result = self.chat.select_multiple(
            message=message,
            choices=choices,
            default=default,
            finish_choice=finish_choice,
            finish_selected_text=finish_selected_text,
            message_end=message_end,
            instruction=instruction,
        )
        _linebreak()
        return result

    def _select_library(self) -> Library | None:
        choices = [
            Choice(title=[*i.title, TextStyle.dim.wrap(i.title_info)], key=i.value, text=i.text)
            for i in Library
        ]
        choices_map = {i.value: i for i in Library}
        selected = self._select(
            "This project uses",
            choices=[*choices, "none of them"],
            default=Library.boto3.value,
        )
        if selected not in choices_map:
            return None
        return choices_map[selected]

    def _select_services(self) -> list[ServiceName]:
        result = self.selected_service_names
        while True:
            self._say(*self._get_services_messages())
            response_choices = list(
                filter(
                    None,
                    [
                        ServiceActions.build if result else None,
                        ServiceActions.add if not self._is_all_selected() else None,
                        ServiceActions.recommended
                        if self.selected_service_names != self.recommended_service_names
                        else None,
                        ServiceActions.remove if result else None,
                        ServiceActions.add_all if not self._is_all_selected() else None,
                        ServiceActions.remove_all if result else None,
                    ],
                )
            )
            select_choices = [
                Choice(
                    title=i.value,
                    key=i.value,
                    text=self._get_build_text() if i == ServiceActions.build else i.value.lower(),
                )
                for i in response_choices
            ]
            response = self._select(
                message="I want to",
                choices=select_choices,
            )
            match response:
                case ServiceActions.build.value:
                    return result
                case ServiceActions.add.value:
                    self._add_service_names()
                    continue
                case ServiceActions.add_all.value:
                    result.clear()
                    result.extend(self.available_service_names)
                    continue
                case ServiceActions.remove.value:
                    self._remove_service_names()
                    continue
                case ServiceActions.remove_all.value:
                    result.clear()
                    continue
                case ServiceActions.recommended.value:
                    result.clear()
                    result.extend(self.recommended_service_names)
                    continue
                case _:
                    return result

    def _get_service_names_args(self) -> list[str]:
        if self._is_all_selected():
            return ["all"]
        return [i.name for i in self.selected_service_names]

    def _find_last_whl(self) -> Path | None:
        prefix = self.library.get_package_prefix()
        packages = list(self.output_path.glob(f"{prefix}*.whl"))
        if not packages:
            return None

        packages.sort(key=lambda x: x.name)
        return packages[-1]

    def _say_commands(self) -> None:
        cmd = " ".join(("uvx", *self.args.to_cmd()))
        lines: list[MessageToken] = [
            "\n\n",
            TextStyle.dim.wrap(f"  # generate {self.library_name} services"),
            "\n  ",
            cmd,
        ]
        found_whl = self._find_last_whl()
        if found_whl:
            lines.extend(
                (
                    "\n\n",
                    TextStyle.dim.wrap(f"  # install with {self.package_manager.value}"),
                    "\n  ",
                    self.package_manager.get_install_cmd(found_whl),
                )
            )
        self._say(lines)

    def _select_package_manager(self) -> PackageManager:
        choices = [
            Choice(title=[f"{i.value:<10}", TextStyle.dim.wrap(i.url)], text=i.value, key=i.value)
            for i in PackageManager
        ]
        choices_map = {i.value: i for i in PackageManager}
        result = self._select(
            "My package manager is",
            choices=[
                *choices,
                Choice(title="not sure", text=["... ", TextStyle.dim.wrap("(maybe pip?)")]),
            ],
        )
        if result not in choices_map:
            self._say(("I see... Let's use ", TextStyle.tag.wrap("pip"), " then."))
            return PackageManager.pip

        return choices_map[result]

    def _select_output_path(self) -> Path:
        response = self.chat.select_output_path(
            message="Save the package to",
            default=DEFAULT_OUTPUT_PATH,
        )
        _linebreak()
        self._respond(
            (
                "I might even add it to ",
                TextStyle.tag.wrap("git"),
                "!",
            )
        )
        return Path(response)

    def _get_cli_namespace(self) -> CLINamespace:
        return CLINamespace(
            log_level=logging.INFO,
            output_path=self.output_path,
            service_names=self._get_service_names_args(),
            build_version="",
            output_types=[OutputType.wheel],
            products=[self.library.product],
            disable_smart_version=False,
            download_static_stubs=True,
        )

    def _do_start_building(self) -> bool:
        result = self.chat.confirm(
            message_yes=(
                "Yes",
                TextStyle.text.wrap(", start building now."),
            ),
            message_no=(
                "No",
                TextStyle.text.wrap(", I will build it later myself."),
            ),
        )
        _linebreak()
        return result

    def _get_build_text(self) -> Message:
        if self._is_all_selected():
            return (
                "continue",
                TextStyle.text.wrap(" with all "),
                TextStyle.tag.wrap("all"),
                TextStyle.text.wrap(" the services. Just in case. I might need them later."),
            )

        if self.selected_service_names == self.recommended_service_names:
            return (
                "continue",
                TextStyle.text.wrap(" with "),
                TextStyle.tag.wrap(f"{len(self.selected_service_names)} recommended"),
                TextStyle.text.wrap(" services."),
            )

        if len(self.selected_service_names) == 1:
            return (
                "continue",
                TextStyle.text.wrap(" with "),
                TextStyle.tag.wrap(self.selected_service_names[0].class_name),
                TextStyle.text.wrap(" service."),
            )

        return (
            "continue",
            TextStyle.text.wrap(" with "),
            TextStyle.tag.wrap(f"{len(self.selected_service_names)} selected"),
            TextStyle.text.wrap(" services."),
        )

    def run(self) -> None:
        """
        Run chat buddy.
        """
        _linebreak()
        self._say(
            ("Hello from ", TextStyle.tag.wrap(PROG_NAME), "!"),
            (
                "It looks like you plan to add ",
                TextStyle.tag.wrap("type checking"),
                " and ",
                TextStyle.tag.wrap("auto-complete"),
                " for ",
                TextStyle.tag.wrap("boto3"),
                ", ",
                TextStyle.tag.wrap("aioboto3"),
                ", or ",
                TextStyle.tag.wrap("aiobotocore"),
                " to your project in ",
                TextStyle.tag.wrap(print_path(self.project_path)),
                " directory.",
            ),
            (
                "You launched me with no ",
                TextStyle.tag.wrap("OUTPUT_PATH"),
                " command-line argument, so I decided to help you a bit.",
            ),
            (
                "By the way, my author did not add any tests for my code.",
                " So, if anything goes wrong, report to ",
                TextStyle.tag.wrap(REPORT_URL),
            ),
        )

        if not self._library:
            self._say(
                (
                    "First of all, what ",
                    TextStyle.tag.wrap("AWS SDK library"),
                    " do you use in this project?",
                ),
            )
            library = self._select_library()

            if not library:
                self._say("No worries, you can always run me again if you start using them.")
                self._finish()
                return

            self._library = library

        self._respond(
            (
                "Now, how can I add ",
                TextStyle.tag.wrap("type checking"),
                " and ",
                TextStyle.tag.wrap("code completion"),
                " for ",
                TextStyle.tag.wrap(self.library_name),
                "?",
            )
        )

        botocore_str = f"botocore {get_botocore_version()}"
        self._say(
            (
                "Oh, it is easy. Just a sec, I am fetching all available services for ",
                TextStyle.tag.wrap(self.library_name),
                " from ",
                TextStyle.tag.wrap(botocore_str),
                " shapes...",
            )
        )
        self.available_service_names = tuple(get_available_service_names())
        self.recommended_service_names = self._get_selected_service_names()
        self.selected_service_names = list(self.recommended_service_names)
        self._say(
            ("Thanks for waiting!",),
            (
                "There are ",
                TextStyle.tag.wrap(len(self.available_service_names)),
                " supported services for ",
                TextStyle.tag.wrap(self.library_name),
                ".",
            ),
            (
                "However, most projects use only ",
                TextStyle.tag.wrap(len(self.selected_service_names)),
                " so I ",
                TextStyle.tag.wrap("recommend"),
                " adding only them.",
            ),
        )
        self.selected_service_names = self._select_services()

        if len(self.selected_service_names) == 1:
            services_str = (
                TextStyle.tag.wrap(self.selected_service_names[0].class_name),
                " service support",
            )
        else:
            services_str = (
                TextStyle.tag.wrap(len(self.selected_service_names)),
                " services support",
            )

        self._say(
            (
                "Great! Let's generate type annotations for ",
                TextStyle.tag.wrap(self.library_name),
                " with ",
                *services_str,
                ".",
            )
        )
        package_name = f"{self.library.get_package_prefix()}_*.whl"

        self._say(
            (
                "I can start building ",
                TextStyle.tag.wrap(package_name),
                " now if you want. Let's start?",
            )
        )
        if not self._do_start_building():
            self._say(
                (
                    "No worries, you can always build ",
                    TextStyle.tag.wrap(self.library.product.value),
                    " later!",
                )
            )
            self._say_commands()
            self._finish()
            return

        self._respond("Go for it!")

        self._say(
            (
                "Almost forgot... Where should I put a generated ",
                TextStyle.tag.wrap(package_name),
                " package?",
            ),
            (
                "I prefer to use ",
                TextStyle.tag.wrap(print_path(self.output_path)),
                " directory, but it is up to you.",
            ),
        )
        self.output_path = self._select_output_path()
        self._say(
            (
                "Good idea! Building to ",
                TextStyle.tag.wrap(print_path(self.output_path)),
                " directory...",
            )
        )

        self.args = self._get_cli_namespace()
        self._run_builder()

    def _run_builder(self) -> None:
        enable_logger()
        self.run_builder(self.args)
        _linebreak()

        found_whl = self._find_last_whl()
        if not found_whl:
            self._say(
                (
                    "All done! But I could not find a built wheel. Something went wrong.",
                    " Please report: ",
                    TextStyle.tag.wrap(REPORT_URL),
                )
            )
            self._say_commands()
            self._finish()
            return

        self._say(
            (
                "I have built ",
                TextStyle.tag.wrap(print_path(found_whl)),
                ", and it is ready to install.",
            ),
            (
                "Let me know which ",
                TextStyle.tag.wrap("package manager"),
                " you use. I can recommend ",
                TextStyle.tag.wrap("uv"),
                ", because it is extremely fast.",
            ),
        )
        self.package_manager = self._select_package_manager()
        self._respond(
            (
                "Sure, I use ",
                TextStyle.tag.wrap(self.package_manager.value),
                ". How to install the generated package?",
            )
        )

        self._say(
            (
                "Use these commands to ",
                TextStyle.tag.wrap("update"),
                " and ",
                TextStyle.tag.wrap("install"),
                " ",
                TextStyle.tag.wrap(self.library.product.value),
                " when you bump ",
                TextStyle.tag.wrap(self.library_name),
                " version:",
            )
        )
        self._say_commands()
        self._finish()

    def _finish(self) -> None:
        self._say(
            (
                "Check ",
                TextStyle.tag.wrap(self.library.documentation_url),
                " documentation. It describes all type annotations for ",
                TextStyle.tag.wrap(self.library_name),
                ".",
            ),
            "Bye-bye! Have a nice day!",
        )



"""
Jinja2 renderer and formatters.

Copyright 2024 Vlad Emelianov
"""

from pathlib import Path

import mdformat

from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.utils.jinja2 import render_jinja2_template
from mypy_boto3_builder.utils.markdown import TableOfContents


def render_jinja2_package_template(template_path: Path, package: Package) -> str:
    """
    Render Jinja2 package template to a string.

    Arguments:
        template_path -- Relative path to template in `TEMPLATES_PATH`
        package -- Service or wrapper package

    Returns:
        A rendered template.
    """
    return render_jinja2_template(
        template_path,
        {
            "package": package,
            "service_name": package.service_name if len(package.service_names) == 1 else None,
        },
    )


def insert_md_toc(text: str) -> str:
    """
    Insert Table of Contents before the first second-level header.
    """
    toc = TableOfContents.parse(text)
    toc_lines = toc.render().splitlines()
    lines = text.splitlines()
    result: list[str] = []
    inserted = False
    for line in lines:
        if not inserted and line.startswith("## "):
            result.extend(toc_lines)
            result.append("")
            inserted = True

        result.append(line)

    if not inserted:
        result.extend(toc_lines)
        result.append("")

    return "\n".join(result)


def format_md(text: str) -> str:
    """
    Format MarkDown with mdformat.
    """
    return mdformat.text(  # type: ignore[partially-unknown]
        text,
        options={
            "wrap": 79,
            "number": True,
        },
    )



"""
Helper for Python import strings.

Copyright 2024 Vlad Emelianov
"""

import copy
import functools
from typing import Self

from mypy_boto3_builder.exceptions import StructureError
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.utils.version import VersionParts


@functools.total_ordering
class ImportRecord:
    """
    Helper for Python import strings.

    Arguments:
        source -- Source of import.
        name -- Import name.
        alias -- Import local name.
        min_version -- Minimum Python version, used for fallback.
        fallback -- Fallback ImportRecord.
    """

    def __init__(
        self,
        source: ImportString,
        name: str = "",
        alias: str = "",
        min_version: VersionParts = (),
        fallback: Self | None = None,
    ) -> None:
        self.source = source
        self.name = name
        self.alias = alias
        self.min_version = min_version
        self.fallback = fallback

    def render_name(self) -> str:
        """
        Get rendered import name.
        """
        if not self.name:
            raise StructureError(f"ImportRecord {self} has no name")

        if self.alias:
            return f"{self.name} as {self.alias}"

        return self.name

    def render(self) -> str:
        """
        Get rendered string.
        """
        if self.name:
            return f"from {self.source} import {self.render_name()}"
        if self.alias:
            return f"import {self.source} as {self.alias}"

        return f"import {self.source}"

    def __str__(self) -> str:
        """
        Render as a valid Python import statement.
        """
        return self.render()

    def __hash__(self) -> int:
        """
        Calculate hash value based on source, name and alias.
        """
        return hash(
            (
                self.source,
                self.name,
                self.alias,
                self.min_version,
                self.fallback,
            ),
        )

    def __eq__(self, other: object) -> bool:
        """
        Whether two import records produce the same render.
        """
        if not isinstance(other, ImportRecord):
            return False

        return str(self) == str(other)

    def __gt__(self, other: Self) -> bool:
        """
        Compare two import records for sorting.

        Emulates `isort` logic.
        """
        if self == other:
            return False

        if self.min_version != other.min_version:
            return self.min_version > other.min_version

        if bool(self.fallback) != bool(other.fallback):
            return bool(self.fallback) > bool(other.fallback)

        if self.source != other.source:
            return self.source > other.source

        return self.name > other.name

    def get_local_name(self) -> str:
        """
        Get local import name.
        """
        return self.alias or self.name or self.source.render()

    def needs_sys_fallback(self) -> bool:
        """
        Whether ImportString requires `sys` module.
        """
        return bool(self.fallback and self.min_version)

    def __copy__(self) -> Self:
        """
        Create a copy of the record.
        """
        return self.__class__(
            source=self.source,
            name=self.name,
            alias=self.alias,
            min_version=self.min_version,
            fallback=self.fallback.copy() if self.fallback else None,
        )

    def copy(self) -> Self:
        """
        Create a copy of the record.
        """
        return copy.copy(self)

    def is_implicit(self) -> bool:
        """
        Whether record can be skipped while rendering: builtins with no fallback.
        """
        return self.source.is_builtins() and not self.fallback



"""
Parser for Boto3 ServiceResource attributes, produces `structures.Attribute`.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.boto3_ports.model import ResourceModel
from mypy_boto3_builder.constants import ATTRIBUTES
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.type_maps.method_type_map import get_method_type_stub


def parse_attributes(
    service_name: ServiceName,
    resource_name: str,
    resource_model: ResourceModel,
    shape_parser: ShapeParser,
) -> list[Attribute]:
    """
    Extract attributes from boto3 resource.

    Arguments:
        resource -- boto3 service resource.

    Returns:
        A list of Attribute structures.
    """
    result: list[Attribute] = []
    if not resource_model.shape:
        return result

    shape = shape_parser.service_model.shape_for(resource_model.shape)
    attributes = resource_model.get_attributes(shape)
    for name, attribute in attributes.items():
        attribute_type = get_method_type_stub(service_name, resource_name, ATTRIBUTES, name)
        if attribute_type is None:
            attribute_shape = attribute[1]
            attribute_type = shape_parser.parse_shape(attribute_shape, is_output_child=True)
        result.append(Attribute(name, attribute_type))

    return result



"""
prompt_toolkit style wrapper.

Copyright 2024 Vlad Emelianov
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from prompt_toolkit.styles import Style

if TYPE_CHECKING:
    from mypy_boto3_builder.chat.type_defs import Message, MessagePair


class TextStyle(Enum):
    """
    prompt_toolkit style wrapper.
    """

    bot = "bot"
    user = "user"
    help = "help"
    text = "text"
    tag = "tag"
    dim = "dim"
    hilight = "hilight"
    qmark = "qmark"
    answer = "answer"
    disabled = "disabled"

    @property
    def style(self) -> str:
        """
        Class usage name.
        """
        return f"class:{self.value}"

    @property
    def css(self) -> str:  # noqa: C901, PLR0911
        """
        CSS properties for class.
        """
        match self:
            case TextStyle.bot:
                return "fg:#762671 bold"
            case TextStyle.user:
                return "fg:#39B54A bold"
            case TextStyle.help:
                return "fg:#006FB8 bold"
            case TextStyle.text:
                return ""
            case TextStyle.tag:
                return "fg:#2CB5E9 bold"
            case TextStyle.dim:
                return "fg:gray"
            case TextStyle.hilight:
                return "fg:#FF9D00 bold"
            case TextStyle.qmark:
                return "fg:#39B54A bold"
            case TextStyle.answer:
                return "fg:#FF9D00 bold"
            case TextStyle.disabled:
                return "hidden"

    @classmethod
    def get_style(cls) -> Style:
        """
        Generate prompt_toolkit.Style.
        """
        return Style([(i.value, i.css) for i in cls])

    def apply(self, message: Message | str) -> tuple[MessagePair, ...]:
        """
        Apply styling for internal usage.
        """
        if not message:
            return ()
        if isinstance(message, str):
            return ((self, message),)

        result: list[MessagePair] = []
        for item in message:
            if isinstance(item, str):
                result.append((self, item))
                continue
            item_style, item_text = item
            if isinstance(item_style, str):
                result.append((self.__class__(item_style.replace("class:", "")), item_text))
                continue
            result.append((item_style, item_text))
        return tuple(result)

    def stylize(self, message: Message | str) -> list[tuple[str, str]]:
        """
        Stylize for prompt_toolkit.
        """
        if not message:
            return []
        if isinstance(message, str):
            return [(self.style, message)]

        result: list[tuple[str, str]] = []
        for item in message:
            if isinstance(item, str):
                result.append((self.style, item))
                continue
            item_style, item_text = item
            if isinstance(item_style, str):
                result.append((item_style, item_text))
                continue
            result.append((item_style.style, item_text))
        return result

    def wrap(self, message: str | float) -> MessagePair:
        """
        Wrap message with style.
        """
        return (self, str(message))



"""
CLI parser.

Copyright 2024 Vlad Emelianov
"""

import argparse
import enum
import logging
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from mypy_boto3_builder.constants import OUTPUT_PATH_SENTINEL, PROG_NAME
from mypy_boto3_builder.enums.output_type import OutputType
from mypy_boto3_builder.enums.product import Product
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.utils.path import print_path
from mypy_boto3_builder.utils.version import get_builder_version


def get_absolute_path(path: str) -> Path:
    """
    Get absolute path from a string.

    Arguments:
        path -- String containing path.

    Returns:
        Absolute path.
    """
    return Path(path).absolute()


class EnumListAction(argparse.Action):
    """
    Argparse action for handling Enums.
    """

    def __init__(
        self,
        *,
        type: type[enum.Enum],  # noqa: A002
        option_strings: Sequence[str],
        dest: str,
        help: str,  # noqa: A002
        default: Sequence[enum.Enum] | None = None,
        nargs: str | None = None,
        required: bool = False,
        **kwargs: str | None,
    ) -> None:
        self._enum_type = type
        self.nargs = nargs or "?"
        choices = tuple(e.value for e in self._enum_type)
        help_str = f"{help}\nChoices: {', '.join(choices)}."

        if default:
            if isinstance(default, enum.Enum):
                help_str += f"\nDefault: {default.value}."
            else:
                help_str += f"\nDefault: {', '.join(e.value for e in default)}."

        super().__init__(
            choices=choices,
            option_strings=option_strings,
            default=default,
            dest=dest,
            nargs=self.nargs,
            type=None,
            help=help_str,
            required=required,
            **kwargs,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        value: str | Sequence[Any] | None,
        _option_string: str | None = None,
    ) -> None:
        """
        Convert value back into an Enum.
        """
        value_list: list[str] = []
        if isinstance(value, str):
            value_list.append(value)
        if isinstance(value, list):
            value_list.extend([i for i in value if isinstance(i, str)])
        enum_values = [self._enum_type(i) for i in value_list]
        result = enum_values if self.nargs != "?" else enum_values[0]
        setattr(namespace, self.dest, result)


@dataclass(kw_only=True, slots=True)
class CLINamespace:
    """
    CLI arguments namespace.
    """

    log_level: int
    output_path: Path
    service_names: list[str]
    build_version: str
    output_types: list[OutputType]
    products: list[Product]
    list_services: bool = False
    partial_overload: bool = False
    skip_published: bool = False
    disable_smart_version: bool = False
    download_static_stubs: bool = True
    raise_interrupt: bool = False

    def to_cmd(self) -> tuple[str, ...]:
        """
        Convert namespace to shlex arguments.
        """
        return tuple(
            filter(
                None,
                (
                    PROG_NAME,
                    print_path(self.output_path),
                    "--no-smart-version" if self.disable_smart_version else None,
                    "--skip-published" if self.skip_published else None,
                    "--partial-overload" if self.partial_overload else None,
                    "--download-static-stubs" if self.download_static_stubs else None,
                    "--list-services" if self.list_services else None,
                    f"--build-version {self.build_version}" if self.build_version else None,
                    f"--product {' '.join(i.value for i in self.products)}"
                    if self.products
                    else None,
                    f"--output-type {' '.join(i.value for i in self.output_types)}"
                    if self.output_types
                    else None,
                    f"--services {' '.join(self.service_names)}" if self.service_names else None,
                    "-d" if self.log_level == logging.DEBUG else None,
                ),
            )
        )


def parse_args(args: Sequence[str]) -> CLINamespace:
    """
    Parse CLI arguments.

    Returns:
        Argument parser.
    """
    version = get_builder_version()

    parser = argparse.ArgumentParser(
        PROG_NAME,
        description="Type annotations generator for boto3, aiobotocore and aioboto3.",
    )
    parser.add_argument("-d", "--debug", action="store_true", help="Show debug messages")
    parser.add_argument(
        "-b",
        "--build-version",
        help="Set custom output version, otherwise smart versioning is used.",
    )
    parser.add_argument("-V", "--version", action="version", version=version)
    parser.add_argument(
        "--product",
        dest="products",
        type=Product,
        action=EnumListAction,
        metavar="PRODUCT",
        nargs="+",
        default=(Product.boto3_stubs, Product.boto3_stubs_services),
        help="Package to generate.",
    )
    parser.add_argument(
        "--skip-published",
        action="store_true",
        help="Skip packages that are already on PyPI.",
    )
    parser.add_argument(
        "--no-smart-version",
        action="store_true",
        help=(
            "Disable version bump based od last PyPI version. "
            "Set this flag to run packages build in offline mode. "
            "skip-published flag is ignored in this case."
        ),
    )
    parser.add_argument(
        "--download-static-stubs",
        action="store_true",
        help="Download static stubs from GitHub repositories instead of using built-in files.",
    )
    parser.add_argument(
        "--panic",
        action="store_true",
        help="Raise exception on logger warning and above",
    )
    parser.add_argument(
        "output_path",
        metavar="OUTPUT_PATH",
        help="Output path",
        type=get_absolute_path,
        default=OUTPUT_PATH_SENTINEL,
        nargs="?",
    )
    parser.add_argument(
        "-s",
        "--services",
        dest="service_names",
        nargs="*",
        metavar="NAME",
        help=(
            "List of AWS services, by default all services are used."
            " Use `updated` to build only services updated in the release."
            " Use `essential` to build only essential services."
            " Use `all` to build all services. (default: all)"
        ),
        default=(ServiceName.ALL,),
    )
    parser.add_argument(
        "--partial-overload",
        action="store_true",
        help="Build client/service overload only for selected services",
    )
    parser.add_argument(
        "--installed",
        action="store_true",
        help=(
            "Generate already installed packages for typings directory."
            " Deprecated, use --output-type=installed instead."
        ),
    )
    parser.add_argument(
        "--output-type",
        type=OutputType,
        action=EnumListAction,
        metavar="OUTPUT_TYPE",
        nargs="*",
        default=(OutputType.package,),
        help="Product output type.",
    )
    parser.add_argument(
        "--list-services",
        action="store_true",
        help="List supported boto3 service names.",
    )
    result = parser.parse_args(args)

    if result.installed:
        result.output_type = [OutputType.installed]

    return CLINamespace(
        log_level=logging.DEBUG if result.debug else logging.INFO,
        output_path=result.output_path,
        service_names=result.service_names,
        products=result.products,
        build_version=result.build_version,
        output_types=result.output_type,
        list_services=result.list_services,
        partial_overload=result.partial_overload,
        skip_published=result.skip_published,
        disable_smart_version=result.no_smart_version,
        download_static_stubs=result.download_static_stubs,
        raise_interrupt=result.debug,
    )



"""
Resource method map for aio libs.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.utils.lookup_dict import LookupDict
from mypy_boto3_builder.utils.type_checks import get_optional

# Mapping to override aio resource methods.
# ServiceName -> Resource name -> Method name -> Method
AIO_RESOURCE_METHOD_MAP: Final[Mapping[ServiceName, Mapping[str, Mapping[str, Method]]]] = {
    ServiceNameCatalog.dynamodb: {
        "Table": {
            "batch_writer": Method(
                name="batch_writer",
                arguments=[
                    Argument.self(),
                    Argument(
                        "overwrite_by_pkeys",
                        get_optional(Type.wrap_list(Type.str)),
                        default=Type.Ellipsis,
                    ),
                    Argument("flush_amount", Type.int, default=Type.Ellipsis),
                    Argument("on_exit_loop_sleep", Type.int, default=Type.Ellipsis),
                ],
                return_type=ExternalImport(
                    Import.aioboto3 + "dynamodb" + "table",
                    "BatchWriter",
                    safe=True,
                ),
            ),
        },
    },
}

_LOOKUP: LookupDict[Method] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in AIO_RESOURCE_METHOD_MAP.items()},
)


def get_aio_resource_method(
    service_name: ServiceName,
    resource_name: str,
    method_name: str,
) -> Method | None:
    """
    Get aio resource method override.

    Arguments:
        service_name -- Service name.
        resource_name -- Parent resource name.
        method_name -- Method name.

    Returns:
        Method structure or None.
    """
    return _LOOKUP.get(service_name.name, resource_name, method_name)



"""
Type annotations for boto3 module.

Copyright 2024 Vlad Emelianov
"""

import logging
from typing import Any

from boto3 import session as session
from boto3.session import Session as Session
from botocore.config import Config
from botocore.session import Session as BotocoreSession

__author__: str = ...
__version__: str = ...

DEFAULT_SESSION: Session | None = ...

def setup_default_session(
    *,
    aws_access_key_id: str | None = ...,
    aws_secret_access_key: str | None = ...,
    aws_session_token: str | None = ...,
    region_name: str | None = ...,
    botocore_session: BotocoreSession | None = ...,
    profile_name: str | None = ...,
) -> None: ...
def set_stream_logger(
    name: str = ...,
    level: int = ...,
    format_string: str | None = ...,
) -> None: ...
def _get_default_session() -> Session: ...

class NullHandler(logging.Handler): ...

def client(
    service_name: str,
    region_name: str | None = ...,
    api_version: str | None = ...,
    use_ssl: bool | None = ...,
    verify: bool | str | None = ...,
    endpoint_url: str | None = ...,
    aws_access_key_id: str | None = ...,
    aws_secret_access_key: str | None = ...,
    aws_session_token: str | None = ...,
    config: Config | None = ...,
) -> Any: ...
def resource(
    service_name: str,
    region_name: str | None = ...,
    api_version: str | None = ...,
    use_ssl: bool | None = ...,
    verify: bool | str | None = ...,
    endpoint_url: str | None = ...,
    aws_access_key_id: str | None = ...,
    aws_secret_access_key: str | None = ...,
    aws_session_token: str | None = ...,
    config: Config | None = ...,
) -> Any: ...



"""
Multiple string utils collection.

Copyright 2024 Vlad Emelianov
"""

import builtins
import datetime
import keyword
import re
import typing
from types import MappingProxyType
from typing import Final, Literal
from unittest.mock import MagicMock

from botocore import xform_name as botocore_xform_name
from botocore.utils import get_service_module_name

from mypy_boto3_builder.constants import DOCSTRING_LINE_LENGTH, DOCSTRING_MAX_LENGTH
from mypy_boto3_builder.exceptions import BuildInternalError, TypeAnnotationError

RESERVED_NAMES: Final = {
    *dir(typing),
    *dir(builtins),
    *keyword.kwlist,
}
AWS_LINK_RE: Final = re.compile(r"`([^`]+\S)\s*<https://(\S+)>`\_*")
REPLACE_DOCSTRING_CHARS: Final = MappingProxyType({"’": "'", "–": "-", " ": " "})  # noqa: RUF001


def get_class_prefix(func_name: str) -> str:
    """
    Get a valid Python class prefix from `func_name`.

    Arguments:
        func_name -- Any string.

    Returns:
        String with a class prefix.
    """
    parts = [capitalize(i) for i in func_name.split("_") if i]
    return "".join(parts)


def get_anchor_link(text: str) -> str:
    """
    Convert header to markdown anchor link.
    """
    return text.strip().replace(" ", "-").replace(".", "").lower()


def is_reserved(word: str) -> bool:
    """
    Check whether variable name conflicts with Python reserved names.
    """
    return word in RESERVED_NAMES


def get_short_docstring(doc: str) -> str:
    """
    Create a short docstring from boto3 documentation.

    Trims docstring to 300 chars.
    Removes double and triple backticks.
    Stops on `**Request syntax**` and `::`.
    Ensures that backticks are closed.
    Replaces `Text <link>` with [Text](link).
    Wraps docstring to 80 chars.
    """
    doc = str(doc)
    if len(doc) > DOCSTRING_MAX_LENGTH:
        doc = f"{doc[: DOCSTRING_MAX_LENGTH - 3]}..."
    result: list[str] = []
    if not doc:
        return ""
    for raw_line in doc.splitlines():
        line = raw_line.strip().removesuffix("::")
        if not line:
            continue
        if line.startswith(":") or line.lower().startswith("**request syntax**"):
            break
        if ". " in line:
            result.append(line.split(". ")[0])
            break
        result.append(line)
        if line.endswith("."):
            break

    result_str = " ".join(result).replace("```", "`").replace("``", "`").replace("\n", " ").strip()
    return clean_artifacts(result_str)


def clean_artifacts(line: str) -> str:
    """
    Remove common artifacts in botocre docs.
    """
    if line.count("`") % 2:
        line = f"{line}`"
    if line and not line.endswith("."):
        line = f"{line}."

    for ch, replacement in REPLACE_DOCSTRING_CHARS.items():
        if ch in line:
            line = line.replace(ch, replacement)

    if "<https:" in line:
        line = AWS_LINK_RE.sub(r"[\g<1>](https://\g<2>)", line)
        # FIXME: temporary fix for pca-connector-ad service
        line = line.replace("https\\:", "https:")
        # FIXME: temporary fix for neptunedata service
        line = line.replace("neptune-db\\:", "neptune-db:")

    return line


def textwrap(text: str, width: int = DOCSTRING_LINE_LENGTH) -> str:
    """
    Wrap text to `width` chars.
    """
    result: list[str] = []
    for raw_line in text.splitlines():
        if len(raw_line) <= width:
            result.append(raw_line)
            continue

        line = raw_line
        while line:
            if len(line) < width:
                result.append(line)
                break
            space_index = line.rfind(" ", 0, width)
            if space_index < 0:
                space_index = line.find(" ", width)

            if space_index < 0:
                result.append(line)
                break

            sub_line = line[:space_index].rstrip()
            next_index = space_index + 1
            line = line[next_index:].lstrip()
            result.append(sub_line)

    return "\n".join(result)


def get_botocore_class_name(metadata: dict[str, str]) -> str:
    """
    Get Botocore class name from Service metadata.
    """
    service_model = MagicMock()
    service_model.service_name = metadata.get("serviceId", "")
    service_model.metadata = metadata
    name = get_service_module_name(service_model)
    return capitalize(name)


def get_type_def_name(*parts: str) -> str:
    """
    Get a valid Python TypeDef class name from `parts`.

    Examples:
        ```python
        get_type_def_name("MyClass", "my_method")  # MyClassMyMethodTypeDef
        ```
    """
    if not parts:
        raise TypeAnnotationError("At least one part is required")

    parts_camelcased = [get_class_prefix(i) for i in parts]
    name = "".join(parts_camelcased)
    return f"{name}TypeDef"


def capitalize(s: str) -> str:
    """
    Capitalize first letter of a string.
    """
    return f"{s[:1].upper()}{s[1:]}"


def xform_name(name: str, sep: str = "_") -> str:
    """
    Convert name to snake_case.

    Arguments:
        name -- Any string.
        sep -- Separator.
    """
    if not sep:
        raise BuildInternalError("Separator is required")
    return botocore_xform_name(name, sep)


def get_pypi_link(package_name: str) -> str:
    """
    Get link to PyPI.
    """
    if not package_name:
        raise BuildInternalError("package_name is required")
    return f"https://pypi.org/project/{package_name}/"


def get_copyright() -> str:
    """
    Get copyright notice.
    """
    now = datetime.datetime.now(datetime.timezone.utc)
    return f"Copyright {now.year} Vlad Emelianov"


def extract_docstring_from_html(html: str) -> str:
    """
    Extract docstring from HTML.
    """
    text = str(html)
    start_index = text.find("<p>")
    if start_index < 0:
        return text
    end_index = text.find("</p>", start_index)
    if end_index < 0:
        return text
    return text[start_index + 3 : end_index].strip()


def get_md_doc_link(
    file: Literal[
        "client",
        "service_resource",
        "waiters",
        "paginators",
        "type_defs",
        "literals",
    ],
    *parts: str,
) -> str:
    """
    Get link to MD docs with anchor.

    Arguments:
        file -- HTML file name
        parts -- Anchor parts
    """
    link = f"./{file}.md"
    if not parts:
        return link
    anchor = "".join([get_anchor_link(part) for part in parts])
    return f"{link}#{anchor}"


def escape_md(value: str) -> str:
    """
    Escape underscore characters.
    """
    return value.replace("_", r"\_")



"""
Path prompt based on questionary.path.

Copyright 2024 Vlad Emelianov
"""

from pathlib import Path

import questionary
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.shortcuts.prompt import CompleteStyle
from prompt_toolkit.validation import ValidationError

from mypy_boto3_builder.chat.prompts.base_prompt import BasePrompt
from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message, MessageToken
from mypy_boto3_builder.chat.utils import as_message


class PathPrompt(BasePrompt[Path]):
    """
    Path prompt based on questionary.path.
    """

    PROMPT_SESSION_MODULE = "questionary.prompts.path.PromptSession"
    HELP_MESSAGE = (
        "Enter a ",
        TextStyle.tag.wrap("path"),
        " to output directory.",
    )

    def __init__(
        self,
        message: Message | str,
        default: Path | None = None,
        instruction: Message | str = "",
    ) -> None:
        super().__init__()
        self.message = as_message(message)
        self.instruction = as_message(instruction) or self.get_help_message()
        self.default = default
        self._result: Path | None = None

    def _validate(self, path: Path | None) -> bool:
        if not path:
            raise ValidationError(0, "Path should not be empty")
        path = Path(path)
        if path.exists() and not path.is_dir():
            raise ValidationError(0, "Path should be a directory")
        return True

    def _patch_answer(self) -> None:
        enter_key_binding = self._get_enter_key_binding()
        orig_set_answer = enter_key_binding.handler

        def set_answer(event: KeyPressEvent) -> None:
            if event.current_buffer.complete_state is None and event.app.current_buffer.validate(
                set_cursor=True
            ):
                self._set_result(Path(event.app.current_buffer.document.text))
            orig_set_answer(event)

        enter_key_binding.handler = set_answer

    def _get_prompt_tokens(self) -> Message:
        tokens: list[MessageToken] = []
        if not self._is_answered():
            tokens.extend(TextStyle.dim.apply(self.instruction))

        tokens.extend(TextStyle.text.apply(self.message))
        return tokens

    def _create_question(self) -> questionary.Question:
        """
        Create an underlying question.
        """
        return questionary.path(
            "",
            default=self.default.as_posix() if self.default is not None else "",
            complete_style=CompleteStyle.READLINE_LIKE,
            style=self.STYLE,
            validate=self._validate,
            only_directories=True,
        )



from typing import TYPE_CHECKING

from mypy_boto3_builder.constants import TEMPLATES_PATH, StaticStubsPath, TemplatePath

if TYPE_CHECKING:
    from pathlib import Path


class TestConstants:
    def test_template_path(self) -> None:
        for key in dir(TemplatePath):
            if key.startswith("_"):
                continue

            path: Path = getattr(TemplatePath, key)
            assert path.is_absolute()
            assert path.exists()
            assert path.relative_to(TEMPLATES_PATH)

    def test_static_stubs_path(self) -> None:
        for key in dir(StaticStubsPath):
            if key.startswith("_"):
                continue
            path: Path = getattr(StaticStubsPath, key)
            assert path.is_absolute()
            assert path.exists()



"""
Protocol for types with children.

Copyright 2024 Vlad Emelianov
"""

from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from typing import Self

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable


class TypeParent(FakeAnnotation, ABC):
    """
    Protocol for types with children.
    """

    @abstractmethod
    def replace_child(self, child: FakeAnnotation, new_child: FakeAnnotation) -> Self:
        """
        Replace child type annotation with a new one.
        """
        ...

    @abstractmethod
    def iterate_direct_type_annotations(self) -> Iterator[FakeAnnotation]:
        """
        Iterate over type annotations directly present in parent.
        """
        ...

    @abstractmethod
    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """
        ...

    def find_type_annotation_parents(self, type_annotation: FakeAnnotation) -> "set[TypeParent]":
        """
        Check recursively if child is present in type def.

        Arguments:
            type_annotation -- Type annotation to search for.
        """
        return (
            self.find_type_annotation_parent_map(
                {type_annotation},
                shallow=False,
            ).get(type_annotation)
            or set()
        )

    def find_type_annotation_parent_map(
        self, type_annotations: Iterable[FakeAnnotation], *, shallow: bool
    ) -> "dict[FakeAnnotation, set[TypeParent]]":
        """
        Check recursively if children are present in type def.

        Arguments:
            type_annotations -- Type annotations to search for.
            shallow -- Skip TypeDefSortable types.

        Can be used for non-overlapping type annotations.
        """
        result: dict[FakeAnnotation, set[TypeParent]] = {}
        for parent in self.find_parents(shallow=shallow):
            for child_type in parent.iterate_direct_type_annotations():
                if child_type not in type_annotations:
                    continue

                if child_type not in result:
                    result[child_type] = set()

                result[child_type].add(parent)

        return result

    def find_parents(self, *, shallow: bool) -> "set[TypeParent]":
        """
        Find all parents recursively including self.

        Arguments:
            shallow -- Skip TypeDefSortable types.
        """
        result: set[TypeParent] = {self}
        stack: list[TypeParent] = [self]

        while stack:
            current = stack.pop()
            for child_type in current.iterate_direct_type_annotations():
                if not isinstance(child_type, TypeParent):
                    continue
                if child_type in result:
                    continue
                if shallow and isinstance(child_type, TypeDefSortable):
                    continue

                result.add(child_type)
                stack.append(child_type)

        return result

    def replace_self_references(self, replacement: FakeAnnotation) -> "set[TypeParent]":
        """
        Replace self references with a new type annotation to avoid recursion.
        """
        """
        Replace self references with a new type annotation to avoid recursion.
        """
        parents = self.find_type_annotation_parents(self)
        for parent in parents:
            parent.replace_child(self, replacement)
        return parents

    def get_sortable_children(self) -> list[TypeDefSortable]:
        """
        Extract required TypeDefSortable list from attributes.
        """
        result: list[TypeDefSortable] = []
        children_types = self.get_children_types()
        for type_annotation in children_types:
            if not isinstance(type_annotation, TypeDefSortable):
                continue
            result.append(type_annotation)

        return result



"""
Main entrypoint for builder.

Copyright 2024 Vlad Emelianov
"""

import datetime
import sys
import warnings
from collections.abc import Iterable, Sequence

from mypy_boto3_builder.chat.chat_buddy import ChatBuddy
from mypy_boto3_builder.cli_parser import CLINamespace, parse_args
from mypy_boto3_builder.constants import (
    BUILDER_REPO_URL,
    OUTPUT_PATH_SENTINEL,
    PACKAGE_NAME,
)
from mypy_boto3_builder.enums.product import Product, ProductLibrary
from mypy_boto3_builder.generators.aioboto3_generator import AioBoto3Generator
from mypy_boto3_builder.generators.aiobotocore_generator import AioBotocoreGenerator
from mypy_boto3_builder.generators.base_generator import BaseGenerator
from mypy_boto3_builder.generators.boto3_generator import Boto3Generator
from mypy_boto3_builder.generators.mypy_boto3_generator import MypyBoto3Generator
from mypy_boto3_builder.generators.types_boto3_generator import TypesBoto3Generator
from mypy_boto3_builder.jinja_manager import JinjaManager
from mypy_boto3_builder.logger import enable_logger, get_logger, setup_logger
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.type_defs import GeneratorKwargs
from mypy_boto3_builder.utils.boto3_utils import get_available_service_names
from mypy_boto3_builder.utils.botocore_changelog import BotocoreChangelog
from mypy_boto3_builder.utils.strings import get_anchor_link, get_copyright, get_md_doc_link
from mypy_boto3_builder.utils.type_checks import (
    is_literal,
    is_type_def,
    is_type_parent,
    is_typed_dict,
    is_union,
)
from mypy_boto3_builder.utils.version import get_builder_version
from mypy_boto3_builder.utils.version_getters import get_botocore_version


def get_selected_service_names(
    selected: Iterable[str],
    available: Iterable[ServiceName],
) -> list[ServiceName]:
    """
    Get a list of selected service names.

    Supports `updated` to select only services updated in currect `boto3` release.
    Supports `all` to select all available service names.

    Arguments:
        selected -- Selected service names as strings.
        available -- All ServiceNames available in current boto3 release.

    Returns:
        A list of selected ServiceNames.
    """
    logger = get_logger()
    available_map = {i.name: i for i in available}
    result: list[ServiceName] = []
    selected_service_names = list(selected)
    botocore_version = get_botocore_version()
    if ServiceName.ALL in selected_service_names:
        return list(available)
    if ServiceName.UPDATED in selected_service_names:
        selected_service_names.remove(ServiceName.UPDATED)
        updated_service_names = BotocoreChangelog().fetch_updated(botocore_version)
        selected_service_names.extend(
            service_name_str
            for service_name_str in updated_service_names
            if service_name_str in available_map
        )
    if ServiceName.ESSENTIAL in selected_service_names:
        selected_service_names.remove(ServiceName.ESSENTIAL)
        selected_service_names.extend(
            service_name_str
            for service_name_str in available_map
            if service_name_str in ServiceName.ESSENTIAL_NAMES
        )

    for service_name_str in selected_service_names:
        if service_name_str not in available_map:
            logger.info(f"Service {service_name_str} is not provided by botocore, skipping")
            continue
        result.append(available_map[service_name_str])

    return result


def get_generator(product: Product, kwargs: GeneratorKwargs) -> BaseGenerator:
    """
    Get Generator class for a product.
    """
    library = product.get_library()
    match library:
        case ProductLibrary.boto3:
            return TypesBoto3Generator(**kwargs)
        case ProductLibrary.mypy_boto3:
            return MypyBoto3Generator(**kwargs)
        case ProductLibrary.boto3_legacy:
            return Boto3Generator(**kwargs)
        case ProductLibrary.aiobotocore:
            return AioBotocoreGenerator(**kwargs)
        case ProductLibrary.aioboto3:
            return AioBoto3Generator(**kwargs)


def generate_product(
    product: Product,
    args: CLINamespace,
    service_names: Sequence[ServiceName],
    main_service_names: Sequence[ServiceName],
) -> None:
    """
    Generate a selected product.

    Arguments:
        product -- Product to generate
        args -- CLI namespace
        service_names -- Selected service names
        main_service_names -- Service names included in main
    """
    generator = get_generator(
        product,
        {
            "product": product,
            "service_names": service_names,
            "main_service_names": main_service_names,
            "config": args,
            "version": args.build_version,
            "cleanup": True,
        },
    )
    generator.generate_product(product.get_type())
    generator.cleanup_temporary_files()


def initialize_jinja_manager() -> None:
    """
    Initialize Jinja manager with globals.
    """
    jinja_manager = JinjaManager()
    jinja_manager.update_globals(
        get_md_doc_link=get_md_doc_link,
        builder_version=get_builder_version(),
        current_year=str(datetime.datetime.now(datetime.timezone.utc).year),
        get_anchor_link=get_anchor_link,
        len=len,
        sorted=sorted,
        repr=repr,
        builder_package_name=PACKAGE_NAME,
        builder_repo_url=BUILDER_REPO_URL,
        copyright=get_copyright(),
        is_typed_dict=is_typed_dict,
        is_union=is_union,
        is_literal=is_literal,
        is_type_def=is_type_def,
        is_type_parent=is_type_parent,
    )


def run(args: CLINamespace) -> None:
    """
    Run builder.
    """
    # FIXME: suppress botocore endpoint warning
    warnings.filterwarnings("ignore", category=FutureWarning, module="botocore.client")

    if args.output_path == OUTPUT_PATH_SENTINEL:
        ChatBuddy(
            run_builder=_run_builder,
            external_args=args,
        ).run()
        return

    _run_builder(args)


def _run_builder(args: CLINamespace) -> None:
    setup_logger(level=args.log_level)
    logger = get_logger()
    available_service_names = get_available_service_names()
    initialize_jinja_manager()
    args.output_path.mkdir(exist_ok=True, parents=True)

    logger.debug(f"{len(available_service_names)} supported botocore services discovered")
    if args.list_services:
        for service_name in available_service_names:
            sys.stdout.write(
                f"- {service_name.name} : {service_name.class_name}"
                f" {service_name.boto3_doc_link}\n",
            )
        return

    service_names = get_selected_service_names(args.service_names, available_service_names)
    main_service_names = service_names if args.partial_overload else available_service_names

    for product in args.products:
        logger.info(f"Generating {product.value} product", tags=product.value)
        generate_product(product, args, service_names, main_service_names)

    logger.debug("Done!")


def main() -> None:
    """
    CLI entrypoint.
    """
    enable_logger()
    args = parse_args(sys.argv[1:])
    try:
        run(args)
    except KeyboardInterrupt:
        if args.raise_interrupt:
            raise
        logger = get_logger()
        logger.warning("Interrupted by user")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Type annotations for aiobotocore.signers module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, Mapping

from aiobotocore.credentials import AioCredentials
from botocore.auth import BaseSigner
from botocore.signers import RequestSigner, S3PostPresigner
from requests.models import Request

class AioRequestSigner(RequestSigner):
    async def handler(
        self, operation_name: Any | None = ..., request: Any | None = ..., **kwargs: Any
    ) -> None: ...
    async def sign(  # type: ignore[override]
        self,
        operation_name: str,
        request: Request,
        region_name: Any | None = ...,
        signing_type: str = ...,
        expires_in: Any | None = ...,
        signing_name: Any | None = ...,
    ) -> None: ...
    async def get_auth_instance(  # type: ignore[override]
        self,
        signing_name: str,
        region_name: str,
        signature_version: Any | None = ...,
        request_credentials: AioCredentials | None = ...,
        **kwargs: Any,
    ) -> BaseSigner: ...
    get_auth: Any
    async def generate_presigned_url(
        self,
        request_dict: Mapping[str, Any],
        operation_name: str,
        expires_in: int = ...,
        region_name: str | None = ...,
        signing_name: str | None = ...,
    ) -> str: ...

def add_generate_db_auth_token(class_attributes: dict[str, Any], **kwargs: Any) -> None: ...
def add_dsql_generate_db_auth_token_methods(
    class_attributes: dict[str, Any], **kwargs: Any
) -> None: ...
async def generate_db_auth_token(
    self: Any, DBHostname: str, Port: int, DBUsername: str, Region: str | None = ...
) -> str: ...
async def dsql_generate_db_connect_auth_token(
    self: Any, Hostname: str, Region: str | None = ..., ExpiresIn: int = ...
) -> str: ...
async def dsql_generate_db_connect_admin_auth_token(
    self: Any, Hostname: str, Region: str | None = ..., ExpiresIn: int = ...
) -> str: ...

class AioS3PostPresigner(S3PostPresigner):
    async def generate_presigned_post(
        self,
        request_dict: Mapping[str, Any],
        fields: Any | None = ...,
        conditions: Any | None = ...,
        expires_in: int = ...,
        region_name: str | None = ...,
    ) -> Any: ...

def add_generate_presigned_url(class_attributes: dict[str, Any], **kwargs: Any) -> None: ...
async def generate_presigned_url(
    self: Any,
    ClientMethod: str,
    Params: Any | None = ...,
    ExpiresIn: int = ...,
    HttpMethod: str | None = ...,
) -> Any: ...
def add_generate_presigned_post(class_attributes: Any, **kwargs: Any) -> None: ...
async def generate_presigned_post(
    self: Any,
    Bucket: str,
    Key: str,
    Fields: Any | None = ...,
    Conditions: Any | None = ...,
    ExpiresIn: int = ...,
) -> Any: ...



"""
Postprocessor for aioboto3 classes and methods.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.postprocessors.aiobotocore import AioBotocorePostprocessor


class AioBoto3Postprocessor(AioBotocorePostprocessor):
    """
    Postprocessor for aioboto3 classes and methods.
    """



"""
Type annotations for aiobotocore.retries.adaptive module.

Copyright 2025 Vlad Emelianov
"""

import logging
from typing import Any

from aiobotocore.client import AioBaseClient
from botocore.retries.adaptive import RateClocker
from botocore.retries.bucket import Clock, TokenBucket
from botocore.retries.standard import ThrottlingErrorDetector
from botocore.retries.throttling import CubicCalculator

logger: logging.Logger = ...

def register_retry_handler(client: AioBaseClient) -> AsyncClientRateLimiter: ...

class AsyncClientRateLimiter:
    def __init__(
        self,
        rate_adjustor: CubicCalculator,
        rate_clocker: RateClocker,
        token_bucket: TokenBucket,
        throttling_detector: ThrottlingErrorDetector,
        clock: Clock,
    ) -> None: ...
    async def on_sending_request(self, request: Any, **kwargs: Any) -> None: ...
    async def on_receiving_response(self, **kwargs: Any) -> None: ...



"""
Wrapper for `typing/typing_extensions.Literal` type annotations like `Literal['a', 'b']`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable
from pathlib import Path
from typing import Self

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.exceptions import TypeAnnotationError
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.internal_import_record import InternalImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_annotation import TypeAnnotation
from mypy_boto3_builder.utils.jinja2 import render_jinja2_template


class TypeLiteral(FakeAnnotation):
    """
    Wrapper for `typing/typing_extensions.Literal` type annotations like `Literal['a', 'b']`.

    Arguments:
        name -- Literal name for non-inline.
        children -- Literal values.
        inline -- Render literal inline.
    """

    def __init__(self, name: str, children: Iterable[str]) -> None:
        self.children: set[str] = set(children)
        self.name: str = name
        if not children:
            raise TypeAnnotationError("Literal should have children")

    def get_sort_key(self) -> str:
        """
        Sort literals by name.
        """
        return self.name

    @property
    def inline(self) -> bool:
        """
        Whether Litereal should be rendered inline.

        1-value literals are rendered inline.
        """
        return len(self.children) == 1

    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if self.inline:
            children = ", ".join([repr(i) for i in sorted(self.children)])
            return f"Literal[{children}]"

        return self.name

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        if self.inline:
            return TypeAnnotation("Literal").get_import_records()

        return {InternalImportRecord(ServiceModuleName.literals, name=self.name)}

    def get_definition_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using Literal.
        """
        return TypeAnnotation("Literal").get_import_records()

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(self.name, self.children)

    def is_literal(self) -> bool:
        """
        Whether type annotation is `Literal`.
        """
        return True

    def add_child(self, child: FakeAnnotation) -> None:
        """
        Disabled method to avoid confusion.
        """
        raise TypeAnnotationError("Use add_literal_child function.")

    def is_same(self, other: Self) -> bool:
        """
        Check if literals have the same children.
        """
        return self.children == other.children

    def get_local_types(self) -> list[FakeAnnotation]:
        """
        Get internal types generated by builder.
        """
        return [self]

    def render_definition(self) -> str:
        """
        Render type annotation definition.
        """
        return render_jinja2_template(Path("common/literal.py.jinja2"), {"literal": self})



"""
Utils for getting underlying package versions.

Copyright 2024 Vlad Emelianov
"""

import contextlib
import functools
import importlib

from botocore import __version__ as botocore_version

from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.utils.pypi_manager import PyPIManager


def get_botocore_version() -> str:
    """
    Get botocore package version.
    """
    return f"{botocore_version}"


def _import_version(module_name: str, version_attr: str = "__version__") -> str:
    with contextlib.suppress(ModuleNotFoundError):
        imported_module = importlib.import_module(module_name)
        return str(getattr(imported_module, version_attr))

    return ""


@functools.cache
def get_boto3_version() -> str:
    """
    Get boto3 package version.
    """
    result = _import_version("boto3")
    if result:
        return result

    result = get_botocore_version()
    get_logger().warning(f"boto3 is not installed, using botocore version: {result}")
    return result


@functools.cache
def get_aiobotocore_version() -> str:
    """
    Get aiobotocore package version.
    """
    result = _import_version("aiobotocore")
    if result:
        return result

    pypi_manager = PyPIManager("aiobotocore")
    result = pypi_manager.get_latest_stable_version()
    get_logger().warning(f"aiobotocore is not installed, using latest version from PyPI: {result}")
    return result


@functools.cache
def get_aioboto3_version() -> str:
    """
    Get aioboto3 package version.
    """
    result = _import_version("aioboto3")
    if result:
        return result

    pypi_manager = PyPIManager("aioboto3")
    result = pypi_manager.get_latest_stable_version()
    get_logger().warning(f"aioboto3 is not installed, using latest version from PyPI: {result}")
    return result



"""
Wrapper for type annotations imported from 3rd party libraries, like `boto3.service.Service`.

Copyright 2024 Vlad Emelianov
"""

import inspect
from typing import Self

from mypy_boto3_builder.exceptions import TypeAnnotationError
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class ExternalImport(FakeAnnotation):
    """
    Wrapper for type annotations imported from 3rd party libraries, like `boto3.service.Service`.

    Arguments:
        source -- Module import string.
        name -- Import name.
        alias -- Import local name.
        safe -- Whether import is wrapped in try-except.
    """

    def __init__(
        self,
        source: ImportString,
        name: str = "",
        alias: str = "",
        *,
        safe: bool = False,
        fallback: ImportRecord | None = None,
    ) -> None:
        self.source: ImportString = source
        self.name: str = name
        self.alias: str = alias
        self.fallback: ImportRecord | None = fallback
        if safe and not self.fallback:
            self.add_fallback(ImportRecord(Import.builtins, "object", self.name))

    @classmethod
    def from_class(cls, obj: type, alias: str = "", *, safe: bool = False) -> Self:
        """
        Create an instance from an imported class.

        Arguments:
            value -- Any Class.
            alias -- Local name.
            safe -- Whether import is wrapped in try-except.
        """
        module = inspect.getmodule(obj)
        if module is None:
            raise TypeAnnotationError(f"Unknown module for {obj}")

        module_name = module.__name__
        return cls(
            source=Import.from_str(module_name),
            name=obj.__name__,
            alias=alias,
            safe=safe,
        )

    @property
    def import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        if self.fallback:
            fallback = self.fallback or ImportRecord(Import.builtins, "object", self.name)
            return ImportRecord(
                self.source,
                self.name,
                self.alias,
                min_version=(),
                fallback=fallback,
            )
        return ImportRecord(
            source=self.source,
            name=self.name,
            alias=self.alias,
            min_version=(),
            fallback=self.fallback,
        )

    def __hash__(self) -> int:
        """
        Calcualte hash value based on import record.
        """
        return hash((self.source, self.name, self.alias, self.fallback))

    def render(self) -> str:
        """
        Get string with local name to use.

        Returns:
            Import record local name.
        """
        return self.import_record.get_local_name()

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        return {self.import_record}

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(self.source, self.name, self.alias, fallback=self.fallback)

    def copy_from(self, other: Self) -> None:
        """
        Copy all fileds from another instance.
        """
        self.source = other.source
        self.name = other.name
        self.fallback = other.fallback

    def add_fallback(self, fallback: ImportRecord | None = None) -> None:
        """
        Add fallback import record.
        """
        self.fallback = fallback or ImportRecord(Import.builtins, "object", self.name)



"""
Chat assistant.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Sequence
from pathlib import Path

from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import FormattedText

from mypy_boto3_builder.chat.choice import Choice
from mypy_boto3_builder.chat.prompts.multiselect_prompt import MultiSelectPrompt
from mypy_boto3_builder.chat.prompts.path_prompt import PathPrompt
from mypy_boto3_builder.chat.prompts.select_prompt import SelectPrompt
from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message
from mypy_boto3_builder.chat.utils import as_message


class Chat:
    """
    Chat assistant.
    """

    BOT_NAME = "NotAI: "
    USER_NAME = "You:   "
    HELP_NAME = "Note:  "

    STYLE = TextStyle.get_style()

    def select_output_path(
        self,
        message: Message | str,
        default: Path | None = None,
    ) -> Path:
        """
        Ask to select an output directory.
        """
        return PathPrompt(
            message=(
                TextStyle.user.wrap(self.USER_NAME),
                *as_message(message),
                " " if message else "",
            ),
            instruction=(
                TextStyle.help.wrap(self.HELP_NAME),
                *PathPrompt.HELP_MESSAGE,
                "\n\n",
            ),
            default=default,
        ).ask()

    def confirm(
        self,
        message: Message | str = "",
        message_end: Message | str = "",
        message_yes: Message | str = "I agree",
        message_no: Message | str = "I disagree",
    ) -> bool:
        """
        Confirm a message.
        """
        choices = [
            Choice(
                title=("Yes ", TextStyle.dim.wrap("(y)")),
                key="yes",
                text=message_yes,
                shortcut_key="y",
            ),
            Choice(
                title=("No ", TextStyle.dim.wrap("(n)")),
                key="no",
                text=message_no,
                shortcut_key="n",
            ),
        ]
        response = self.select(
            message=message,
            message_end=message_end,
            choices=choices,
        )
        return response == "yes"

    def select(
        self,
        message: Message | str,
        choices: Sequence[Choice | str],
        *,
        default: Choice | str | None = None,
        message_end: Message | str = "",
        instruction: Message | str = "",
    ) -> str:
        """
        Select one item from list.
        """
        result = SelectPrompt(
            message=(
                TextStyle.user.wrap(self.USER_NAME),
                *as_message(message),
                " " if message else "",
            ),
            choices=choices,
            instruction=(
                TextStyle.help.wrap(self.HELP_NAME),
                *as_message(instruction or SelectPrompt.HELP_MESSAGE),
                "\n\n",
            ),
            message_end=message_end,
            default=default,
        ).ask()
        return result[0].key

    def select_multiple(
        self,
        message: Message | str,
        choices: Sequence[Choice | str],
        *,
        default: Choice | str | None = None,
        message_end: Message | str = "",
        instruction: Message | str = "",
        finish_choice: Choice | str = "Go back",
        finish_selected_text: Message | str | None = None,
    ) -> list[str]:
        """
        Select multiple items from list.
        """
        prompt = MultiSelectPrompt(
            message=(
                TextStyle.user.wrap(self.USER_NAME),
                *as_message(message),
                " " if message else "",
            ),
            choices=choices,
            instruction=(
                TextStyle.help.wrap(self.HELP_NAME),
                *MultiSelectPrompt.HELP_MESSAGE,
                "\n\n",
            ),
            finish_choice=finish_choice,
            finish_selected_text=finish_selected_text,
            message_end=message_end,
            default=default,
        )
        prompt.instruction = (
            TextStyle.help.wrap(self.HELP_NAME),
            *prompt.get_help_message(),
            " " if instruction else "",
            *as_message(instruction),
            "\n\n",
        )
        result = prompt.ask()
        return [i.key for i in result]

    def respond(self, message: Message | str) -> None:
        """
        Respond as a User.
        """
        result = FormattedText(
            [*TextStyle.user.stylize(self.USER_NAME), *TextStyle.text.stylize(message)]
        )
        print_formatted_text(result, style=self.STYLE, end="\n\n")

    def say(self, message: Message | str) -> None:
        """
        Say as a Bot.
        """
        formatted_text = FormattedText(
            [*TextStyle.bot.stylize(self.BOT_NAME), *TextStyle.text.stylize(message)]
        )
        print_formatted_text(formatted_text, style=self.STYLE, end="\n\n")



from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog


class TestServiceName:
    def test_init(self) -> None:
        service_name = ServiceName("my-service", "MyService")
        assert service_name.name == "my-service"
        assert service_name.import_name == "my_service"
        assert service_name.class_name == "MyService"
        assert service_name.boto3_version == "latest"
        assert service_name.boto3_name == "my-service"
        assert service_name.extras_name == "my-service"
        assert (
            service_name.boto3_doc_link
            == "https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/my-service.html#myservice"
        )
        assert (
            service_name.boto3_doc_link_parent
            == "https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/my-service"
        )

    def test_special_name(self) -> None:
        assert ServiceName("lambda", "MyService").import_name == "lambda_"

    def test_is_essential(self) -> None:
        assert not ServiceName("my-service", "MyService").is_essential()
        assert ServiceName("lambda", "MyService").is_essential()


class TestServiceNameCatalog:
    def test_add(self) -> None:
        assert ServiceNameCatalog.add("test", "Test").name == "test"
        assert ServiceNameCatalog.add("test", "Test").name == "test"



"""
Wrapper for Python import strings.

Copyright 2024 Vlad Emelianov
"""

import functools
from typing import Final, Self

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.exceptions import BuildInternalError, StructureError
from mypy_boto3_builder.import_helpers.import_parent import ImportParent
from mypy_boto3_builder.package_data import (
    Boto3StubsPackageData,
    TypesAioBotocorePackageData,
    TypesBoto3PackageData,
)


@functools.total_ordering
class ImportString:
    """
    Wrapper for Python import strings.

    Arguments:
        parent -- Parent module name
        parts -- Other import parts

    Examples::

        import_string = ImportString("my", "name")

        str(import_string)
        'my.name'

        import_string.render()
        'my.name'
    """

    def __init__(self, parent: str, *parts: str) -> None:
        all_parts = (parent, *parts)
        if not parent and not parts:
            raise StructureError("ImportString cannot be empty")
        has_not_empty_part = False
        for part in all_parts:
            if "." in part:
                raise StructureError(f"ImportString parts are not splitted correctly: {all_parts}")
            if part:
                has_not_empty_part = True
            elif has_not_empty_part:
                raise StructureError(
                    f"ImportString cannot have empty parts after parents: {all_parts}",
                )

        self.parts: Final[tuple[str, ...]] = tuple(all_parts)

    def __str__(self) -> str:
        """
        Render as a part of a valid Python import statement.
        """
        return self.render()

    def __hash__(self) -> int:
        """
        Calculate hash value based on all parts.
        """
        return hash(self.parts)

    def __eq__(self, other: object) -> bool:
        """
        Whether import strings produce the same render.
        """
        if not isinstance(other, ImportString):
            raise BuildInternalError(f"{other} is not ImportString")

        return self.parts == other.parts

    def __gt__(self, other: Self) -> bool:
        """
        Compare import strings for sorting.

        Emulates `isort` logic.
        """
        if self == other:
            return False

        self_is_local, other_is_local = self.is_local(), other.is_local()
        if self_is_local != other_is_local:
            return self_is_local > other_is_local

        self_is_third_party, other_is_third_party = self.is_third_party(), other.is_third_party()
        if self_is_third_party != other_is_third_party:
            return self_is_third_party > other_is_third_party

        return self.parts > other.parts

    def __add__(self, other: Self | str) -> Self:
        """
        Create a new import string by adding another import string parts to the end.
        """
        other_import_string = other if isinstance(other, ImportString) else ImportString(other)
        return self.__class__(*self.parts, *other_import_string.parts)

    def render(self) -> str:
        """
        Render to string.

        Returns:
            Ready to use import string.
        """
        return ".".join(self.parts)

    @property
    def parent(self) -> str:
        """
        Get first import string part or `builtins`.
        """
        return self.parts[0]

    def is_local(self) -> bool:
        """
        Whether import is from local module.
        """
        if self.parent.startswith(TypesBoto3PackageData.service_prefix):
            return True

        if self.parent.startswith(Boto3StubsPackageData.service_prefix):
            return True

        if self.parent.startswith(TypesAioBotocorePackageData.service_prefix):
            return True

        return self.is_type_defs()

    def is_builtins(self) -> bool:
        """
        Whether import is from Python `builtins` module.
        """
        return ImportParent.is_builtins(self.parent)

    def is_type_defs(self) -> bool:
        """
        Whether import is from `type_defs` module.
        """
        if not self.parts:
            return False
        return self.parts[-1] == ServiceModuleName.type_defs.value

    def is_third_party(self) -> bool:
        """
        Whether import is from 3rd party module.
        """
        return ImportParent.is_third_party(self.parent)

    def startswith(self, other: "ImportString") -> bool:
        """
        Whether import string starts with another import string.
        """
        return self.parts[: len(other.parts)] == other.parts



"""
Type annotations for aiobotocore.paginate module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, AsyncIterator, Generic, TypeVar

from botocore.paginate import PageIterator, Paginator

_R = TypeVar("_R")

class AioPageIterator(PageIterator[_R], Generic[_R]):
    def __aiter__(self) -> AsyncIterator[_R]: ...
    async def __anext__(self) -> _R: ...
    def result_key_iters(self) -> Any: ...
    async def build_full_result(self) -> dict[str, Any]: ...  # type: ignore[override]
    def search(self, expression: str) -> AsyncIterator[_R]: ...  # type: ignore[override]

class AioPaginator(Paginator[_R], Generic[_R]):
    PAGE_ITERATOR_CLS: type[AioPageIterator[Any]]  # type: ignore[override]

class ResultKeyIterator(Generic[_R]):
    result_key: str
    def __init__(self, pages_iterator: PageIterator[_R], result_key: str) -> None: ...
    def __aiter__(self) -> AsyncIterator[_R]: ...
    async def __anext__(self) -> _R: ...



"""
Class or module attribute.

Copyright 2024 Vlad Emelianov
"""

import copy
from collections.abc import Generator
from typing import Literal, Self

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant

_TypeIgnore = Literal["override"] | None


class Attribute:
    """
    Class or module attribute.

    Arguments:
        name -- Attribute name.
        type -- Attribute type annotation.
        value -- Attribute value.
        type_ignore -- Add type: ignore[<type_ignore>] comment.
        is_reference -- Whether the attribute parsed from references.
        is_identifier -- Whether the attribute parsed from identifiers.
        is_collection -- Whether the attribute parsed from collections.
    """

    def __init__(
        self,
        name: str,
        type_annotation: FakeAnnotation,
        value: TypeConstant | None = None,
        *,
        type_ignore: _TypeIgnore = None,
        is_reference: bool = False,
        is_identifier: bool = False,
        is_collection: bool = False,
    ) -> None:
        self.name = name
        self.type_annotation = type_annotation
        self.value = value
        self.type_ignore: _TypeIgnore = type_ignore
        self.is_reference = is_reference
        self.is_identifier = is_identifier
        self.is_collection = is_collection

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Iterate over all type annotations used.

        Yields:
            Type annotation.
        """
        return self.type_annotation.iterate_types()

    def render(self) -> str:
        """
        Render to a string.
        """
        line = f"{self.name}: {self.type_annotation.render()}"
        if self.type_ignore:
            return f"{line}  # type: ignore[{self.type_ignore}]"
        return line

    def is_autoload_property(self) -> bool:
        """
        Whether the attribute is an autoload property.
        """
        if self.name == "meta":
            return False

        return not self.is_reference and not self.is_identifier and not self.is_collection

    def copy(self) -> Self:
        """
        Deep copy attribute.
        """
        return copy.copy(self)

    def __copy__(self) -> Self:
        """
        Deep copy attribute.
        """
        return self.__class__(
            name=self.name,
            type_annotation=self.type_annotation.copy(),
            value=self.value.copy() if self.value else None,
            type_ignore=self.type_ignore,
            is_reference=self.is_reference,
            is_identifier=self.is_identifier,
            is_collection=self.is_collection,
        )



"""
Types for botocore shape files.

Copyright 2024 Vlad Emelianov
"""

from typing import Any, NotRequired, TypedDict


class PaginatorShape(TypedDict):
    """
    Paginator shape.
    """

    input_token: list[str] | str
    limit_key: NotRequired[str]


class PaginatorsShape(TypedDict):
    """
    Paginators shape.
    """

    pagination: dict[str, PaginatorShape]


class IdentifierShape(TypedDict):
    """
    Identifier shape.
    """

    name: NotRequired[str]
    target: NotRequired[str]
    source: NotRequired[str]
    type: NotRequired[str]


class SubResourceShape(TypedDict):
    """
    SubResource shape.
    """

    type: str
    identifiers: NotRequired[list[IdentifierShape]]
    path: NotRequired[str]


class HasShape(TypedDict):
    """
    Has shape.
    """

    resource: NotRequired[SubResourceShape]


class ParamShape(TypedDict):
    """
    Param shape.
    """

    target: str
    source: str
    value: str | int | float | None


class RequestShape(TypedDict):
    """
    Request shape.
    """

    operation: str
    params: NotRequired[list[ParamShape]]


class ActionShape(TypedDict):
    """
    Action shape.
    """

    request: RequestShape
    resource: SubResourceShape


class WaiterShape(TypedDict):
    """
    Waiter shape.
    """

    operation: str


class WaitersShape(TypedDict):
    """
    Waiters shape.
    """

    waiters: dict[str, WaiterShape]


class ResourceShape(TypedDict):
    """
    Resource shape.
    """

    shape: NotRequired[str]
    identifiers: NotRequired[list[IdentifierShape]]
    actions: NotRequired[dict[str, Any]]
    waiters: NotRequired[dict[str, WaiterShape]]
    has: NotRequired[dict[str, HasShape]]


class ResourcesShape(TypedDict):
    """
    Resources shape.
    """

    service: ResourceShape
    resources: NotRequired[dict[str, ResourceShape]]



"""
Helper for Python import strings without parent module name.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.import_helpers.import_record import ImportRecord


class InternalImportRecord(ImportRecord):
    """
    Helper for Python import strings without parent module name.

    Arguments:
        service_module_name -- Service module name.
        name -- Import name.
        alias -- Import local name.
    """

    def __init__(
        self,
        service_module_name: ServiceModuleName,
        name: str = "",
        alias: str = "",
    ) -> None:
        source = Import.local(service_module_name.name)
        super().__init__(source, name=name, alias=alias)



"""
Type annotations for aiobotocore.httpchecksum module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, Mapping, TypeVar

from aiobotocore.response import StreamingBody
from aiohttp import StreamReader
from botocore.awsrequest import AWSHTTPResponse, AWSRequest
from botocore.httpchecksum import AwsChunkedWrapper, BaseChecksum
from botocore.model import OperationModel

_R = TypeVar("_R")

class AioAwsChunkedWrapper(AwsChunkedWrapper):
    async def _make_chunk(self) -> bytes: ...
    def __aiter__(self: _R) -> _R: ...
    async def __anext__(self) -> bytes: ...

class StreamingChecksumBody(StreamingBody):
    def __init__(
        self,
        raw_stream: StreamReader,
        content_length: str,
        checksum: BaseChecksum,
        expected: BaseChecksum,
    ) -> None: ...
    async def read(self, amt: int | None = ...) -> bytes: ...

async def handle_checksum_body(
    http_response: AWSHTTPResponse,
    response: dict[str, Any],
    context: Mapping[str, Any],
    operation_model: OperationModel,
) -> None: ...
def apply_request_checksum(request: AWSRequest) -> None: ...



"""
Class method.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterator

from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.function import Function


class Method(Function):
    """
    Class method.
    """

    def get_self_argument(self) -> Argument | None:
        """
        Get `self` or `cls` argument.
        """
        if not self.arguments:
            return None
        first_argument = self.arguments[0]
        if first_argument.name in Argument.FIRST_NAMES:
            return first_argument

        return None

    def iterate_call_arguments(self) -> Iterator[Argument]:
        """
        Iterate over arguments that are used in method call.
        """
        self_argument = self.get_self_argument()
        for argument in super().iterate_call_arguments():
            if argument == self_argument:
                continue
            yield argument

    def iterate_packed_arguments(self) -> Iterator[Argument]:
        """
        Iterate over packed arguments for KW-only methods.
        """
        packed_arguments = super().iterate_packed_arguments()
        if not self.is_kw_only() or not self.has_request_type_annotation():
            yield from packed_arguments
            return

        self_argument = self.get_self_argument()
        if self_argument:
            yield self_argument

        yield from packed_arguments

    def has_arguments(self) -> bool:
        """
        Whether method has arguments.
        """
        self_argument = self.get_self_argument()
        if not self_argument:
            return super().has_arguments()

        return len(self.arguments) > 1



"""
Type annotations for aiobotocore.handlers module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, Mapping

from botocore.signers import RequestSigner
from requests.models import Response

async def check_for_200_error(response: Response, **kwargs: Any) -> None: ...
async def inject_presigned_url_ec2(
    params: Mapping[str, Any], request_signer: RequestSigner, model: Any, **kwargs: Any
) -> None: ...
async def inject_presigned_url_rds(
    params: Mapping[str, Any], request_signer: RequestSigner, model: Any, **kwargs: Any
) -> None: ...
async def parse_get_bucket_location(
    parsed: dict[str, Any], http_response: Response, **kwargs: Any
) -> None: ...



"""
Type checking utils.

Copyright 2024 Vlad Emelianov
"""

from typing import TypeGuard

from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_parent import TypeParent
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.type_annotations.type_union import TypeUnion


def is_typed_dict(annotation: FakeAnnotation) -> TypeGuard[TypeTypedDict]:
    """
    Whether type annotation is TypedDict.
    """
    return isinstance(annotation, TypeTypedDict)


def is_union(annotation: FakeAnnotation) -> TypeGuard[TypeUnion]:
    """
    Whether type annotation is a Union.
    """
    return isinstance(annotation, TypeUnion)


def is_literal(annotation: FakeAnnotation) -> TypeGuard[TypeLiteral]:
    """
    Whether type annotation is a literal.
    """
    return isinstance(annotation, TypeLiteral)


def is_type_def(annotation: FakeAnnotation) -> TypeGuard[TypeDefSortable]:
    """
    Whether type annotation is a named TypeDefSortable.
    """
    if is_typed_dict(annotation):
        return True

    if is_union(annotation):
        return annotation.is_named()

    return False


def is_type_parent(annotation: FakeAnnotation) -> TypeGuard[TypeParent]:
    """
    Whether type annotation is a TypeParent.
    """
    return isinstance(annotation, TypeParent)


def is_external_import(annotation: FakeAnnotation) -> TypeGuard[ExternalImport]:
    """
    Whether type annotation is a ExternalImport.
    """
    return isinstance(annotation, ExternalImport)


def is_type_subscript(annotation: FakeAnnotation) -> TypeGuard[TypeSubscript]:
    """
    Whether type annotation is a TypeSubscript.
    """
    return isinstance(annotation, TypeSubscript)


def get_optional(wrapped: FakeAnnotation) -> TypeUnion:
    """
    Get Optional type annotation.
    """
    if is_union(wrapped):
        result = wrapped.copy()
        result.add_child(Type.none)
        return result

    return TypeUnion([wrapped, Type.none])



"""
Wrapper for TypedDict maps.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator

from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.utils.type_def_sorter import TypeDefSorter


class TypedDictMap(dict[str, TypeTypedDict]):
    """
    Wrapper for TypedDict maps.
    """

    def add(self, item: TypeTypedDict) -> None:
        """
        Add new item.
        """
        self[item.name] = item

    def iterate_by_name(self, name: str) -> Generator[TypeTypedDict]:
        """
        Iterate over items matched by real dict name.
        """
        items = list(self.values())
        for item in items:
            if item.name == name:
                yield item

    def rename(self, item: TypeTypedDict, new_name: str) -> None:
        """
        Rename item and change mapping.
        """
        for key, value in list(self.items()):
            if value == item:
                del self[key]

        item.name = new_name
        self[new_name] = item

    def get_sorted_names(self) -> list[str]:
        """
        Get real TypedDict names topologically sorted.
        """
        sorted_values = TypeDefSorter(set(self.values())).sort()
        allowed_names = {i.name for i in self.values()}
        return [i.name for i in sorted_values if i.name in allowed_names]



"""
Type annotations for aiobotocore.configprovider module.

Copyright 2025 Vlad Emelianov
"""

from botocore.configprovider import ConfigValueStore, SmartDefaultsConfigStoreFactory

class AioSmartDefaultsConfigStoreFactory(SmartDefaultsConfigStoreFactory):
    async def merge_smart_defaults(  # type: ignore[override]
        self,
        config_store: ConfigValueStore,
        mode: str,
        region_name: str,
    ) -> None: ...
    async def resolve_auto_mode(self, region_name: str) -> str: ...  # type: ignore[override]



"""
Collection of TypedDicts added by boto3.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_typed_dict import TypedDictAttribute, TypeTypedDict
from mypy_boto3_builder.type_annotations.type_union import TypeUnion

WaiterConfigTypeDef = TypeTypedDict(
    "WaiterConfigTypeDef",
    (
        TypedDictAttribute("Delay", Type.int, required=False),
        TypedDictAttribute("MaxAttempts", Type.int, required=False),
    ),
)

PaginatorConfigTypeDef = TypeTypedDict(
    "PaginatorConfigTypeDef",
    (
        TypedDictAttribute("MaxItems", Type.int, required=False),
        TypedDictAttribute("PageSize", Type.int, required=False),
        TypedDictAttribute("StartingToken", Type.str, required=False),
    ),
)

ResponseMetadataTypeDef = TypeTypedDict(
    "ResponseMetadataTypeDef",
    (
        TypedDictAttribute("RequestId", Type.str, required=True),
        TypedDictAttribute("HostId", Type.str, required=False),
        TypedDictAttribute("HTTPStatusCode", Type.int, required=True),
        TypedDictAttribute("HTTPHeaders", Type.DictStrStr, required=True),
        TypedDictAttribute("RetryAttempts", Type.int, required=True),
    ),
)

# s3
CopySourceTypeDef = TypeTypedDict(
    "CopySourceTypeDef",
    (
        TypedDictAttribute("Bucket", Type.str, required=True),
        TypedDictAttribute("Key", Type.str, required=True),
        TypedDictAttribute("VersionId", Type.str, required=False),
    ),
)

# ec2
TagTypeDef = TypeTypedDict(
    "TagTypeDef",
    (
        TypedDictAttribute("Key", Type.str, required=False),
        TypedDictAttribute("Value", Type.str, required=False),
    ),
)


# dynamodb
EmptyResponseMetadataTypeDef = TypeTypedDict(
    "EmptyResponseMetadataTypeDef",
    (TypedDictAttribute("ResponseMetadata", ResponseMetadataTypeDef, required=True),),
)

# dynamodb
AttributeValueTypeDef: TypeTypedDict = TypeTypedDict(
    "AttributeValueTypeDef",
    (
        TypedDictAttribute("S", Type.str, required=False),
        TypedDictAttribute("N", Type.str, required=False),
        TypedDictAttribute("B", Type.bytes, required=False),
        TypedDictAttribute("SS", TypeSubscript(Type.Sequence, [Type.str]), required=False),
        TypedDictAttribute("NS", TypeSubscript(Type.Sequence, [Type.str]), required=False),
        TypedDictAttribute("BS", TypeSubscript(Type.Sequence, [Type.bytes]), required=False),
        TypedDictAttribute("M", Type.MappingStrAny, required=False),
        TypedDictAttribute("L", Type.SequenceAny, required=False),
        TypedDictAttribute("NULL", Type.bool, required=False),
        TypedDictAttribute("BOOL", Type.bool, required=False),
    ),
)

# cloudformation
GetTemplateOutputTypeDef = TypeTypedDict(
    "GetTemplateOutputTypeDef",
    (
        TypedDictAttribute("TemplateBody", Type.DictStrAny, required=True),
        TypedDictAttribute(
            "StagesAvailable",
            TypeSubscript(
                Type.List,
                [InternalImport(name="TemplateStageType", module_name=ServiceModuleName.literals)],
            ),
            required=True,
        ),
        TypedDictAttribute("ResponseMetadata", ResponseMetadataTypeDef, required=True),
    ),
)

# iam
# FIXME: https://github.com/boto/botocore/issues/2992
PolicyDocumentStatementTypeDef = TypeTypedDict(
    "PolicyDocumentStatementTypeDef",
    (
        TypedDictAttribute("Effect", Type.str, required=True),
        TypedDictAttribute(
            "Resource",
            TypeUnion((Type.str, TypeSubscript(Type.List, [Type.str]))),
            required=True,
        ),
        TypedDictAttribute("Sid", Type.str, required=True),
        TypedDictAttribute(
            "Action",
            TypeUnion((Type.str, TypeSubscript(Type.List, [Type.str]))),
            required=True,
        ),
    ),
)

PolicyDocumentDictTypeDef = TypeTypedDict(
    "PolicyDocumentDictTypeDef",
    (
        TypedDictAttribute("Version", Type.str, required=True),
        TypedDictAttribute(
            "Statement",
            TypeSubscript(Type.List, [PolicyDocumentStatementTypeDef]),
            required=True,
        ),
    ),
)


# cloudwatch events
CloudwatchEventStateTypeDef = TypeTypedDict(
    "CloudwatchEventStateTypeDef",
    (
        TypedDictAttribute("reason", Type.str, required=False),
        TypedDictAttribute("reasonData", Type.str, required=False),
        TypedDictAttribute("timestamp", Type.str, required=True),
        TypedDictAttribute("value", Type.str, required=True),
        TypedDictAttribute("actionsSuppressedBy", Type.str, required=False),
        TypedDictAttribute("actionsSuppressedReason", Type.str, required=False),
    ),
)

CloudwatchEventMetricStatsMetricTypeDef = TypeTypedDict(
    "CloudwatchEventMetricStatsMetricTypeDef",
    (
        TypedDictAttribute("metricName", Type.str, required=True),
        TypedDictAttribute("namespace", Type.str, required=True),
        TypedDictAttribute("dimensions", Type.DictStrStr, required=True),
    ),
)

CloudwatchEventMetricStatsTypeDef = TypeTypedDict(
    "CloudwatchEventMetricStatsTypeDef",
    (
        TypedDictAttribute("metric", CloudwatchEventMetricStatsMetricTypeDef, required=False),
        TypedDictAttribute("period", Type.str, required=True),
        TypedDictAttribute("stat", Type.str, required=True),
    ),
)

CloudwatchEventMetricTypeDef = TypeTypedDict(
    "CloudwatchEventMetricTypeDef",
    (
        TypedDictAttribute("id", Type.str, required=True),
        TypedDictAttribute("metricStat", CloudwatchEventMetricStatsTypeDef, required=False),
        TypedDictAttribute("returnData", Type.bool, required=True),
        TypedDictAttribute("expression", Type.str, required=False),
        TypedDictAttribute("label", Type.str, required=False),
        TypedDictAttribute("period", Type.int, required=False),
    ),
)

CloudwatchEventConfigurationTypeDef = TypeTypedDict(
    "CloudwatchEventDetailConfigurationTypeDef",
    (
        TypedDictAttribute("id", Type.str, required=False),
        TypedDictAttribute("description", Type.str, required=False),
        TypedDictAttribute(
            "metrics",
            TypeSubscript(Type.List, [CloudwatchEventMetricTypeDef]),
            required=False,
        ),
        TypedDictAttribute("actionsSuppressor", Type.str, required=False),
        TypedDictAttribute("actionsSuppressorWaitPeriod", Type.int, required=False),
        TypedDictAttribute("actionsSuppressorExtensionPeriod", Type.int, required=False),
        TypedDictAttribute("threshold", Type.int, required=False),
        TypedDictAttribute("evaluationPeriods", Type.int, required=False),
        TypedDictAttribute("alarmRule", Type.str, required=False),
        TypedDictAttribute("alarmName", Type.str, required=False),
        TypedDictAttribute("treatMissingData", Type.str, required=False),
        TypedDictAttribute("comparisonOperator", Type.str, required=False),
        TypedDictAttribute("timestamp", Type.str, required=False),
        TypedDictAttribute("actionsEnabled", Type.bool, required=False),
        TypedDictAttribute("okActions", TypeSubscript(Type.List, [Type.str]), required=False),
        TypedDictAttribute("alarmActions", TypeSubscript(Type.List, [Type.str]), required=False),
        TypedDictAttribute(
            "insufficientDataActions",
            TypeSubscript(Type.List, [Type.str]),
            required=False,
        ),
    ),
)

CloudwatchEventDetailTypeDef = TypeTypedDict(
    "CloudwatchEventDetailTypeDef",
    (
        TypedDictAttribute("alarmName", Type.str, required=True),
        TypedDictAttribute("operation", Type.str, required=False),
        TypedDictAttribute("configuration", CloudwatchEventConfigurationTypeDef, required=False),
        TypedDictAttribute(
            "previousConfiguration",
            CloudwatchEventConfigurationTypeDef,
            required=False,
        ),
        TypedDictAttribute("previousState", CloudwatchEventStateTypeDef, required=False),
        TypedDictAttribute("state", CloudwatchEventStateTypeDef, required=True),
    ),
)

CloudwatchEventTypeDef = TypeTypedDict(
    "CloudwatchEventTypeDef",
    (
        TypedDictAttribute("version", Type.str, required=True),
        TypedDictAttribute("id", Type.str, required=True),
        TypedDictAttribute("detail-type", Type.str, required=True),
        TypedDictAttribute("source", Type.str, required=True),
        TypedDictAttribute("account", Type.str, required=True),
        TypedDictAttribute("time", Type.str, required=True),
        TypedDictAttribute("region", Type.str, required=True),
        TypedDictAttribute("resources", TypeSubscript(Type.List, [Type.str]), required=True),
        TypedDictAttribute("detail", CloudwatchEventDetailTypeDef, required=True),
    ),
)



"""
Type annotations for aioboto3.resources.action module.

Copyright 2025 Vlad Emelianov
"""

import logging
from typing import Any

from boto3.resources.action import ServiceAction, WaiterAction
from boto3.resources.base import ServiceResource
from boto3.resources.factory import ResourceFactory
from boto3.resources.model import Action
from boto3.utils import ServiceContext

logger: logging.Logger

class AIOServiceAction(ServiceAction):
    def __init__(
        self,
        action_model: Action,
        factory: ResourceFactory | None = ...,
        service_context: ServiceContext | None = ...,
    ) -> None: ...
    async def __call__(  # type: ignore[override]
        self, parent: ServiceResource, *args: Any, **kwargs: Any
    ) -> ServiceResource | list[ServiceResource] | dict[str, Any]: ...

class AioBatchAction(ServiceAction):
    async def __call__(  # type: ignore[override]
        self, parent: ServiceResource, *args: Any, **kwargs: Any
    ) -> list[dict[str, Any]]: ...

class AIOWaiterAction(WaiterAction):
    async def __call__(  # type: ignore[override]
        self, parent: ServiceResource, *args: Any, **kwargs: Any
    ) -> None: ...



"""
Type annotations for aioboto3.s3.inject module.

Copyright 2025 Vlad Emelianov
"""

import logging
import sys
from abc import abstractmethod
from typing import Any, BinaryIO, Callable, Sequence

from boto3.s3.transfer import TransferConfig as S3TransferConfig

if sys.version_info >= (3, 12):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

logger: logging.Logger

TransferCallback: TypeAlias = Callable[[int], None]

class _AsyncBinaryIO:
    @abstractmethod
    async def seek(self, offset: int, whence: int = 0) -> int: ...
    @abstractmethod
    async def write(self, s: bytes | bytearray) -> int: ...

AnyFileObject: TypeAlias = _AsyncBinaryIO | BinaryIO

def inject_s3_transfer_methods(class_attributes: Sequence[Any], **kwargs: Any) -> None: ...
def inject_object_summary_methods(class_attributes: Sequence[Any], **kwargs: Any) -> None: ...
def inject_bucket_methods(class_attributes: Sequence[Any], **kwargs: Any) -> None: ...
async def object_summary_load(self: Any, *args: Any, **kwargs: Any) -> None: ...
async def download_file(
    self: Any,
    Bucket: str,
    Key: str,
    Filename: str,
    ExtraArgs: list[str] | None = ...,
    Callback: TransferCallback | None = ...,
    Config: S3TransferConfig | None = ...,
) -> None: ...
async def download_fileobj(
    self: Any,
    Bucket: str,
    Key: str,
    Fileobj: AnyFileObject,
    ExtraArgs: dict[str, Any] | None = ...,
    Callback: TransferCallback | None = ...,
    Config: S3TransferConfig | None = ...,
) -> None: ...
async def upload_fileobj(
    self: Any,
    Fileobj: AnyFileObject,
    Bucket: str,
    Key: str,
    ExtraArgs: dict[str, Any] | None = ...,
    Callback: TransferCallback | None = ...,
    Config: S3TransferConfig | None = ...,
    Processing: Callable[[bytes], bytes] | None = ...,
) -> Any: ...
async def upload_file(
    self: Any,
    Filename: str,
    Bucket: str,
    Key: str,
    ExtraArgs: dict[str, Any] | None = ...,
    Callback: TransferCallback | None = ...,
    Config: S3TransferConfig | None = ...,
) -> None: ...
async def copy(
    self: Any,
    CopySource: Any,
    Bucket: str,
    Key: str,
    ExtraArgs: dict[str, Any] | None = ...,
    Callback: TransferCallback | None = ...,
    SourceClient: Any | None = ...,
    Config: S3TransferConfig | None = ...,
) -> None: ...
async def bucket_load(self: Any, *args: Any, **kwargs: Any) -> None: ...



"""
Boto3 Client.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterator
from typing import Final

from botocore.client import BaseClient
from botocore.errorfactory import BaseClientExceptions

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.utils.type_checks import is_typed_dict


class Client(ClassRecord):
    """
    Service Client.
    """

    _OWN_METHOD_NAMES: Final = {
        "get_waiter",
        "get_paginator",
        "exceptions",
    }

    def __init__(self, name: str, service_name: ServiceName) -> None:
        super().__init__(name=name, bases=[ExternalImport.from_class(BaseClient)])
        self.service_name = service_name
        self.exceptions_class = ClassRecord(
            name="Exceptions",
            bases=(ExternalImport.from_class(BaseClientExceptions),),
        )

    def __hash__(self) -> int:
        """
        Calculate hash from client service name.
        """
        return hash(self.service_name)

    @property
    def alias_name(self) -> str:
        """
        Class alias name for safe import.
        """
        return "Client"

    @property
    def boto3_doc_link(self) -> str:
        """
        List to boto3 docs page.
        """
        return (
            f"{self.service_name.boto3_doc_link_parent}.html#{self.service_name.class_name}.Client"
        )

    def get_all_names(self) -> list[str]:
        """
        Get a list of names for `__all__` statement.
        """
        return [self.name]

    @property
    def own_methods(self) -> Iterator[Method]:
        """
        Get a list of auto-generated methods.
        """
        for method in self.methods:
            if method.name not in self._OWN_METHOD_NAMES:
                yield method

    def get_exceptions_property(self) -> Method:
        """
        Generate Client exceptions property.
        """
        return Method(
            name="exceptions",
            decorators=[ExternalImport.from_class(property)],
            arguments=[
                Argument.self(),
            ],
            return_type=InternalImport(
                name=self.exceptions_class.name,
                module_name=ServiceModuleName.client,
                service_name=self.service_name,
            ),
            docstring=f"{self.name} exceptions.",
            boto3_doc_link=self.boto3_doc_link,
        )

    def get_example_method(self) -> Method | None:
        """
        Get a nice method with return TypedDict for documentation.
        """
        for method in self.methods:
            if not method.has_request_type_annotation():
                continue
            if not method.return_type:
                continue
            if not is_typed_dict(method.return_type):
                continue

            return method

        return None
